
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.contentstransformation;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.zipentrytransformation.RenameZipEntryTransformer;
import de.unkrig.commons.file.zipentrytransformation.SelectiveZipEntryTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntriesTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryContentsTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryTransformerUtil;
import de.unkrig.commons.io.IoUtil;
import de.unkrig.commons.lang.protocol.Predicate;
import de.unkrig.commons.text.pattern.Glob;

/**
 * Utility class for {@link ContentsTransformer}.
 */
public final
class ContentsTransformerUtil {

    private
    ContentsTransformerUtil() {}

    /**
     * @return A {@link ContentsTransformer} which {@link IoUtil#copy(InputStream, OutputStream)}s all data
     *         unmodified.
     */
    public static ContentsTransformer
    copy() { return ContentsTransformerUtil.COPY; }

    private static final ContentsTransformer COPY = new ContentsTransformer() {

        @Override public void
        transform(String name, InputStream is, OutputStream os) throws IOException {
            IoUtil.copy(is, os);
        }
    };

    /**
     * Creates a contents transformer that transforms the contents (optionally looking into ZIP and GZIP format
     * files, recursively, renaming and/or removing entries on-the-fly) through the given contents transformer.
     * <pre>
     *                           |
     *    +---------+            v
     *    |         |          ZipCT
     *    |         v      entry| |dflt
     *    |       ZipCT         | |
     *    |   entry|  |dflt     | |
     *    |        |  +----------------+
     *    |        v            | |    |
     *    |     SelZET<---------+ |    |
     *    |      |  |             |    |
     *    |      |  v             |    |
     *    |      | NOP            |    |
     *    |      v                |    |
     *    |  RenameZET            |    |
     *    |      |                |    |
     *    |      v                |    |
     *    |    ZECT               |    |
     *    |      |                |    |
     *    |      v                |    |
     *    |    GzipCT          GzipCT  |
     *    |     |  |            |  |   |
     *    |     v  v            |  |   |
     *    +-----+<-+<-----------+  |   |
     *                             v   v
     *                           defaultCT
     * </pre>
     */
    public static ContentsTransformer
    newZipContentsTransformer(
        boolean                       lookIntoZip,
        boolean                       lookIntoNestedZip,
        boolean                       lookIntoGzip,
        boolean                       lookIntoNestedGzip,
        Predicate<String>             zipEntryRemoval,
        Glob                          zipEntryRenaming,
        ZipEntriesTransformer         zipEntriesTransformer,
        boolean                       keepOriginals,
        ContentsTransformer           defaultContentsTransformer,
        ExceptionHandler<IOException> exceptionHandler
    ) {

        final ZipEntryTransformer[] zipEntryTransformer = new ZipEntryTransformer[1];

        ContentsTransformer contentsTransformer;
        if (lookIntoZip && lookIntoNestedZip) {
            ZipContentsTransformer zipContentsTransformer = new ZipContentsTransformer(
                zipEntriesTransformer,
                new ZipEntryTransformer() {

                    @Override public void
                    transform(ZipEntry entry, String path, InputStream is, ZipOutputStream zos) throws IOException {
                        zipEntryTransformer[0].transform(entry, path, is, zos);
                    }
                },
                defaultContentsTransformer,
                exceptionHandler
            );

            contentsTransformer = (
                lookIntoGzip
                ? new GzipContentsTransformer(zipContentsTransformer, defaultContentsTransformer)
                : defaultContentsTransformer
            );

            zipEntryTransformer[0] = new ZipEntryContentsTransformer(
                lookIntoNestedGzip
                ? new GzipContentsTransformer(zipContentsTransformer, zipContentsTransformer)
                : zipContentsTransformer
            );
        } else {
            contentsTransformer = (
                lookIntoGzip
                ? new GzipContentsTransformer(defaultContentsTransformer, defaultContentsTransformer)
                : defaultContentsTransformer
            );

            zipEntryTransformer[0] = new ZipEntryContentsTransformer(defaultContentsTransformer);
        }

        if (lookIntoZip) {

            zipEntryTransformer[0] = new RenameZipEntryTransformer(zipEntryRenaming, zipEntryTransformer[0]);

            zipEntryTransformer[0] = new SelectiveZipEntryTransformer(
                zipEntryRemoval,
                ZipEntryTransformerUtil.nop("Removing entry ''{0}''"),
                zipEntryTransformer[0]
            );

            contentsTransformer = new ZipContentsTransformer(
                zipEntriesTransformer,  // zipEntriesTransformer
                zipEntryTransformer[0], // zipEntryTransformer
                contentsTransformer     // defaultContentsTransformer
            );
        }

        return contentsTransformer;
    }
}
