
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.contentstransformation;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.zipentrytransformation.ZipEntriesTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryTransformer;
import de.unkrig.commons.io.UnclosableInputStream;
import de.unkrig.commons.io.UnclosableOutputStream;
import de.unkrig.commons.lang.ExceptionUtil;

/**
 * A {@link ContentsTransformer} that delegates contents transformation to a {@link ZipEntriesTransformer} and a {@link
 * ZipEntryTransformer}, or a delegate {@link ContentsTransformer}one of two delegates, depending on whether the input
 * is in ZIP format or not.
 *
 * @see #transform(String, InputStream, OutputStream)
 */
public
class ZipContentsTransformer implements ContentsTransformer {

    private final ZipEntriesTransformer         zipEntriesTransformer;
    private final ZipEntryTransformer           zipEntryTransformer;
    private final ContentsTransformer           defaultContentsTransformer;
    private final ExceptionHandler<IOException> exceptionHandler;

    /**
     * @see #transform(String, InputStream, OutputStream)
     */
    public
    ZipContentsTransformer(
        ZipEntriesTransformer zipEntriesTransformer,
        ZipEntryTransformer   zipEntryTransformer,
        ContentsTransformer   defaultContentsTransformer
    ) {
        this.zipEntriesTransformer      = zipEntriesTransformer;
        this.zipEntryTransformer        = zipEntryTransformer;
        this.defaultContentsTransformer = defaultContentsTransformer;
        this.exceptionHandler           = ExceptionHandler.defaultHandler();
    }

    /**
     * @see #transform(String, InputStream, OutputStream)
     */
    public
    ZipContentsTransformer(
        ZipEntriesTransformer         zipEntriesTransformer,
        ZipEntryTransformer           zipEntryTransformer,
        ContentsTransformer           defaultContentsTransformer,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        this.zipEntriesTransformer      = zipEntriesTransformer;
        this.zipEntryTransformer        = zipEntryTransformer;
        this.defaultContentsTransformer = defaultContentsTransformer;
        this.exceptionHandler           = exceptionHandler;
    }

    /**
     * If the contents of {@code is} is not in ZIP format, then it is transformed through {@code
     * defaultContentsTransformer}. Otherwise, it is opened as a {@link ZipInputStream}, a {@link ZipOutputStream} is
     * created on {@code os}, and for each entry on the {@link ZipInputStream}, an entry with the same name is created
     * on the {@link ZipOutputStream}, and then this method is called recursively on the entry contents, and the
     * transformed output is stored in the {@link ZipOutputStream}.
     */
    @Override public void
    transform(String path, InputStream is, OutputStream os) throws IOException {

        // Make the stream resettable.
        if (!is.markSupported()) is = new BufferedInputStream(is);
        is.mark(4);

        // Check if the stream contents starts with the magic number.
        boolean isZip;
        {
            byte[] buf = new byte[4];
            try {
                isZip = (is.read(buf) == 4 && buf[3] == 0x04 && buf[2] == 0x03 && buf[1] == 0x4B && buf[0] == 0x50);
            } catch (IOException ioe) {
                this.exceptionHandler.handle(ExceptionUtil.wrap("Reading the first few bytes", ioe));
                return;
            } catch (RuntimeException re) {
                this.exceptionHandler.handle(ExceptionUtil.wrap("Reading the first few bytes", re));
                return;
            }
            try { is.reset(); } catch (Exception e) {}
        }

        if (!isZip) {
            this.defaultContentsTransformer.transform(path, is, os);
            return;
        }

        // We need to close the ZIS later (to release the resources allocated by the ZIP inflater), but the contract
        // of 'transform()' requires that the underlying 'is' must not be closed.
        ZipInputStream zis = new ZipInputStream(new UnclosableInputStream(is));
        try {
            ZipOutputStream zos = new ZipOutputStream(new UnclosableOutputStream(os));
            try {

                ZipEntry entry;
                try {
                    entry = zis.getNextEntry();
                } catch (ZipException e) {

                    // In rare cases, we get a 'ZipException: invalid entry CRC (expected 0x0 but got 0xfde2c118)' -
                    // which obviously means that we hit a binary file with the correct magic number, but an
                    // otherwise invalid first entry.
                    is.reset();
                    this.defaultContentsTransformer.transform(path, is, zos);
                    return;
                } catch (IOException ioe) {
                    throw ExceptionUtil.wrap("Reading first entry", ioe);
                } catch (RuntimeException re) {
                    throw ExceptionUtil.wrap("Reading first entry", re);
                }

                if (entry == null) {

                    // This should not happen, because it indicates a wrong magic number, and the magic number
                    // has been checked before.
                    is.reset();
                    this.defaultContentsTransformer.transform(path, is, os);
                    return;
                }

                String prefix = path + '!';

                do {
                    String entryName = path + '!' + entry.getName();

                    this.zipEntriesTransformer.visitEntry(prefix, entry, zis, this.zipEntryTransformer, zos);

                    try {
                        entry = zis.getNextEntry();
                    } catch (IOException ioe) {
                        this.exceptionHandler.handle(
                            ExceptionUtil.wrap("Reading next entry after '" + entryName + "'", ioe)
                        );
                        return;
                    } catch (RuntimeException re) {
                        this.exceptionHandler.handle(
                            ExceptionUtil.wrap("Reading next entry after '" + entryName + "'", re)
                        );
                        return;
                    }
                } while (entry != null);

                this.zipEntriesTransformer.visitEnd(prefix, zos);

                zos.close();
            } finally {
                try { zos.close(); } catch (Exception e) {}
            }
            zis.close();
        } finally {
            try { zis.close(); } catch (Exception e) {}
        }
    }
}
