
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.fileprocessing;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.contentsprocessing.ContentsProcessor;
import de.unkrig.commons.file.contentsprocessing.GzipContentsProcessor;
import de.unkrig.commons.file.contentsprocessing.SelectiveContentsProcessor;
import de.unkrig.commons.file.contentsprocessing.ZipContentsProcessor;
import de.unkrig.commons.file.zipentryprocessing.ZipEntryContentsProcessor;
import de.unkrig.commons.lang.ExceptionUtil;
import de.unkrig.commons.lang.protocol.Duplexer;
import de.unkrig.commons.lang.protocol.HardReference;
import de.unkrig.commons.lang.protocol.Predicate;
import de.unkrig.commons.nullanalysis.Nullable;

/**
 * Various file processing utility methods.
 */
public final
class FileProcessingUtil {

    private
    FileProcessingUtil() {}

    /**
     * @see #process(List, FileProcessor, ExceptionHandler)
     */
    public static void
    process(List<File> files, FileProcessor fileProcessor) throws IOException {
        FileProcessingUtil.process(files, fileProcessor, ExceptionHandler.<IOException>defaultHandler());
    }

    /**
     * Invokes the {@code fileProcessor} for each of the {@code files}.
     *
     * @param exceptionHandler Called if the processing of one of the {@code inputFiles} throws an {@link IOException}
     *                         or a {@link RuntimeException}
     */
    public static void
    process(List<File> files, FileProcessor fileProcessor, ExceptionHandler<IOException> exceptionHandler)
    throws IOException {
        for (File file : files) {
            try {
                fileProcessor.process(file.getPath(), file);
            } catch (IOException ioe) {
                exceptionHandler.handle(ExceptionUtil.wrap("Processing '" + file + "'", ioe));
            } catch (RuntimeException re) {
                exceptionHandler.handle(ExceptionUtil.wrap("Processing '" + file + "'", re));
            }
        }
    }

    /**
     * The "swiss army knife" for scanning directories, zip and gzip files.
     * <p>
     * If all boolean parameters are {@code false}, then a {@link FileProcessor} is returned which invokes the {@code
     * delegate} for all files who's pathes match the {@code pathPredicate}. (This will only work for flat
     * files.)
     * <p>
     * The boolean parameters determine where files should be checked for actually being directories ({@code
     * lookIntoDirectories}, {@code lookIntoNestedDirectories}), or having ZIP format {@code lookIntoZip}, {@code
     * lookIntoNestedZip}), or having GZIP format {@code lookIntoGzip}).
     */
    @SuppressWarnings("unchecked") public static FileProcessor
    newFileProcessor(
        boolean                             lookIntoDirectories,
        boolean                             lookIntoNestedDirectories,
        boolean                             lookIntoZip,
        final boolean                       lookIntoNestedZip,
        final boolean                       lookIntoGzip,
        final Predicate<String>             pathPredicate,
        ContentsProcessor                   delegate,
        final ExceptionHandler<IOException> exceptionHandler
    ) {
        return FileProcessingUtil.newDirectoryTreeProcessor(
            lookIntoDirectories,
            lookIntoNestedDirectories,
            pathPredicate,
            FileProcessingUtil.newZipFileProcessor(
                lookIntoZip,
                pathPredicate,
                FileProcessingUtil.combine(
                    new SelectiveContentsProcessor(pathPredicate, delegate, ContentsProcessor.NOP),
                    FileProcessingUtil.newZipContentsDuplexer(lookIntoNestedZip, pathPredicate),
                    FileProcessingUtil.newGzipContentsDuplexer(lookIntoGzip)
                ),
                new SelectiveFileProcessor(pathPredicate, new FileContentsProcessor(delegate), FileProcessor.NOP)
            ),
            exceptionHandler
        );
    }

    /**
     * @return A {@link Duplexer} that produces a {@link ContentsProcessor} which uses {@code delegate1} for
     *         subcontents in GZIP format, and {@code delegate2} otherwise
     */
    public static Duplexer<ContentsProcessor>
    newGzipContentsDuplexer(final boolean lookIntoGzip) {
        return new Duplexer<ContentsProcessor>() {

            @Override public ContentsProcessor
            produce(ContentsProcessor subcontentsDelegate, ContentsProcessor plainContentsDelegate) {
                return FileProcessingUtil.newGzipContentsProcessor(
                    lookIntoGzip,
                    subcontentsDelegate,
                    plainContentsDelegate
                );
            }
        };
    }

    /**
     * @return A {@link Duplexer} that produces a {@link ContentsProcessor} which uses {@code delegate1} for
     *         the entries of contents in ZIP format, and {@code delegate2} otherwise
     */
    public static Duplexer<ContentsProcessor>
    newZipContentsDuplexer(final boolean lookIntoZip, final Predicate<String> pathPredicate) {
        return new Duplexer<ContentsProcessor>() {

            @Override public ContentsProcessor
            produce(ContentsProcessor subcontentsDelegate, ContentsProcessor plainContentsDelegate) {
                return FileProcessingUtil.newZipContentsProcessor(
                    lookIntoZip,
                    pathPredicate,
                    subcontentsDelegate,
                    plainContentsDelegate
                );
            }
        };
    }

    /**
     * Wraps the given {@code delegate} contents processor such that it processes contained 'subcontents', e.g.
     * entries of a ZIP archive, GZIPped contents, TARred contents, etc.
     * <p>
     * Handles also <i>nested</i> subcontents in any combination, e.g. zipped ZIP archives, gzipped TAR files etc.
     * <p>
     * The {@code duplexer}s must produce {@link ContentsProcessor}s which invoke {@code delegate1} for contents
     * that contains subcontents, and {@code delegate2} if it contains plain contents.
     * <p>
     * E.g. if {@code duplexers} produces TWO content processors 'CP1' and 'CP2':
     * <pre>
     *      Any contents
     *           |
     *           v
     *    +----->+
     *    |      v
     *    |     CP1
     *    |   d1| |d2
     *    +<----+ |
     *    ^       v
     *    |      CP2
     *    |    d1| |d2
     *    +<-----+ |
     *             v
     *          delegate
     * </pre>
     *
     * @param duplexers Produces the {@link ContentsProcessor}s for the subcontents; must inv
     */
    public static ContentsProcessor
    combine(ContentsProcessor delegate, Duplexer<ContentsProcessor>... duplexers) {
        final HardReference<ContentsProcessor> loopback   = new HardReference<ContentsProcessor>();
        ContentsProcessor                      toLoopback = new ContentsProcessor() {

            @SuppressWarnings("null") @Override public void
            process(String name, InputStream is) throws IOException { loopback.get().process(name, is); }
        };

        for (Duplexer<ContentsProcessor> duplexer : duplexers) {
            delegate = duplexer.produce(toLoopback, delegate);
        }

        loopback.set(delegate);
        return delegate;
    }

    /**
     * If all boolean parameters are {@code false}, then a {@link FileProcessor} is returned which invokes the {@code
     * delegate} contents processor for all files who's pathes match the {@code pathPredicate}. (This will only work
     * for flat files.)
     * <p>
     * The boolean parameters determine where files should be checked for actually have ZIP format {@code lookIntoZip},
     * {@code lookIntoNestedZip}).
     * <pre>
     *               Any file
     *                   |
     *                   v
     *              lookIntoZip?
     *               y|      |n
     *                v      |
     *           SelFP:name! |
     *              y|    |n |
     *               v    |  |
     *             ZipFP  |  |
     *            y|   |n |  |
     *             |   |  |  |
     *           ZECP  |  |  |
     *             |   |  |  |
     *             v   v  v  v
     *      delegate1 delegate2
     * </pre>
     */
    public static FileProcessor
    newZipFileProcessor(
        boolean           lookIntoZip,
        Predicate<String> pathPredicate,
        ContentsProcessor delegate1,
        FileProcessor     delegate2
    ) {
        if (!lookIntoZip) return delegate2;

        return new SelectiveFileProcessor(
            FileProcessingUtil.stringPredicateSuffixed(pathPredicate, "!"),
            new ZipFileProcessor(new ZipEntryContentsProcessor(delegate1), delegate2),
            delegate2
        );
    }

    /**
     * If {@code lookIntoZip} is {@code false}, then a {@link ContentsProcessor} is returned which invokes the {@code
     * plainContentsDelegate} contents processor. Otherwise, and if the file is in ZIP format, the {@code
     * subcontentsDelegate} is invoked with the contents of each ZIP entry.
     * <pre>
     *                Any content
     *                     |
     *                     v
     *                lookIntoZip?
     *                y|        |n
     *                 v        |
     *            SelCP:name!   |
     *             y|     |n    v
     *              v     +---->+
     *            ZipCP         |
     *            y| |n         v
     *             | +--------->+
     *             v            |
     *           ZECP           |
     *             |            |
     *             v            v
     * subcontentsDelegate plainContentsDelegate
     * </pre>
     */
    private static ContentsProcessor
    newZipContentsProcessor(
        boolean           lookIntoZip,
        Predicate<String> pathPredicate,
        ContentsProcessor subcontentsDelegate,
        ContentsProcessor plainContentsDelegate
    ) {
        if (!lookIntoZip) return plainContentsDelegate;

        return new SelectiveContentsProcessor(
            FileProcessingUtil.stringPredicateSuffixed(pathPredicate, "!"),
            new ZipContentsProcessor(new ZipEntryContentsProcessor(subcontentsDelegate), plainContentsDelegate),
            plainContentsDelegate
        );
    }

    /**
     * If {@code lookIntoGzip} is {@code false}, then a {@link ContentsProcessor} is returned which invokes the {@code
     * plainContentsDelegate}. Otherwise, if the contents is in GZIP format, it is gunzipped and the {@code
     * subcontentsDelegate} invoked.
     * <pre>
     *                 Contents
     *                     |
     *                     v
     *                lookIntoGzip?
     *                  y|     |n
     *                   v     |
     *                 GzipCP  |
     *                y|    |n |
     *                 v    v  v
     * subcontentsDelegate plainContentsDelegate
     * </pre>
     */
    private static ContentsProcessor
    newGzipContentsProcessor(
        boolean           lookIntoGzip,
        ContentsProcessor subcontentsDelegate,
        ContentsProcessor plainContentsDelegate
    ) {
        return (
            lookIntoGzip
            ? new GzipContentsProcessor(subcontentsDelegate, plainContentsDelegate)
            : plainContentsDelegate
        );
    }

    /**
     * If {@code lookIntoDirectories} is {@code false}, then the {@code delegate} is returned. Otherwise each file is
     * check if it actually is a director<, and, if so, the {@code delegate} is invoked for each member of the
     * directory. If {@code lookIntoNestedDirectories} is also {@code true}, then this process is executed recursively,
     * i.e. the {@code delegate} will only be called for non-directory files.
     * <pre>
     *                Any file
     *                    |
     *                    v
     *           lookIntoDirectories?
     *                y|       |n
     *                 v       |
     *  +--------->SelFP:name/ |
     *  |             y|    |n |
     *  |              v    |  |
     *  |            DirFP  |  |
     *  |           y|   |n |  |
     *  |            v   |  |  |
     *  | lookIntoNested |  |  |
     *  |  Directories?  |  |  |
     *  |   |y       n|  |  |  |
     *  +---+         v  v  v  v
     *                 delegate
     * </pre>
     */
    public static FileProcessor
    newDirectoryTreeProcessor(
        boolean                       lookIntoDirectories,
        boolean                       lookIntoNestedDirectories,
        Predicate<String>             pathPredicate,
        FileProcessor                 delegate,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        if (!lookIntoDirectories) return delegate;

        if (!lookIntoNestedDirectories) {
            return new SelectiveFileProcessor(
                FileProcessingUtil.stringPredicateSuffixed(pathPredicate, "/"),
                new DirectoryProcessor(delegate, delegate, exceptionHandler),
                delegate
            );
        }

        final HardReference<FileProcessor> loopback   = new HardReference<FileProcessor>();
        FileProcessor                      toLoopback = new FileProcessor() {

            @SuppressWarnings("null") @Override public void
            process(String path, File file)
            throws FileNotFoundException, IOException { loopback.get().process(path, file); }
        };

        FileProcessor fileProcessor2 = new DirectoryProcessor(toLoopback, delegate, exceptionHandler);

        FileProcessor result = new SelectiveFileProcessor(
            FileProcessingUtil.stringPredicateSuffixed(pathPredicate, "/"),
            fileProcessor2,
            delegate
        );
        loopback.set(result);
        return result;
    }

    private static Predicate<String>
    stringPredicateSuffixed(final Predicate<String> delegate, final String suffix) {
        return new Predicate<String>() {

            @Override public boolean
            evaluate(@Nullable String subject) { return delegate.evaluate(subject + suffix); }

            @Override public String
            toString() { return delegate.toString() + "[" + suffix + "]"; }
        };
    }
}
