
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.filetransformation;

import java.io.File;
import java.io.IOException;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.FileUtil;

/**
 * @see #transform(String, File, File)
 */
public
class DirectoryTransformer implements FileTransformer {

    private final FileTransformer               memberTransformer;
    private final FileTransformer               defaultFileTransformer;
    private final ExceptionHandler<IOException> exceptionHandler;
    private final DirectoryMembersTransformer   directoryMembersTransformer;

    /**
     * @see #transform(String, File, File)
     */
    public
    DirectoryTransformer(
        FileTransformer               defaultFileTransformer,
        DirectoryMembersTransformer   directoryMembersTransformer,
        FileTransformer               memberTransformer,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        this.defaultFileTransformer      = defaultFileTransformer;
        this.directoryMembersTransformer = directoryMembersTransformer;
        this.memberTransformer           = memberTransformer;
        this.exceptionHandler            = exceptionHandler;
    }

    /**
     * If {@code in} is not a directory, then the {@code defaultFileTransformer} is invoked.
     * <p>
     * Otherwise, {@code out} is taken as a directory (created if missing), and {@link
     * #DirectoryTransformer(FileTransformer, DirectoryMembersTransformer, FileTransformer, ExceptionHandler)
     * memberTransformer}{@code .transform()} is called for each member of {@code in}. If that throws an {@link
     * IOException} or a {@link RuntimeException}, then that exception is caught, and {@link
     * #DirectoryTransformer(FileTransformer, DirectoryMembersTransformer, FileTransformer, ExceptionHandler)
     * exceptionHandler}{@code .handle()} is called. If that method completes normally (i.e. it doesn't rethrow its
     * given exception), then processing continues with the next member of directory {@code in}.
     */
    @Override public void
    transform(String path, File in, File out) throws IOException {

        // Delegate to {@code regularFileTransformer} if {@code in} is not a directory.
        if (!in.isDirectory()) {
            this.defaultFileTransformer.transform(path, in, out);
            return;
        }

        // Create output directory if necessary.
        if (!in.equals(out) && !out.isDirectory()) {
            if (!out.mkdirs()) throw new IOException("Could not create directory '" + out + "'");
        }

        // Transform all directory members.
        for (File member : in.listFiles()) {
            try {
                this.directoryMembersTransformer.visitMember(
                    path + File.separatorChar + member.getName(),
                    member,
                    new File(out, member.getName()),
                    this.memberTransformer
                );
            } catch (IOException ioe) {
                if (!in.equals(out)) FileUtil.attemptToDeleteRecursively(out);
                this.exceptionHandler.handle(ioe);
            } catch (RuntimeException re) {
                if (!in.equals(out)) FileUtil.attemptToDeleteRecursively(out);
                this.exceptionHandler.handle(re);
            }
        }

        this.directoryMembersTransformer.visitEnd(in, out);
    }
}
