
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.filetransformation;

import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * A {@link FileTransformer} that ignores the input file (on out-of-place transformations), or deletes the input
 * file (on in-place-transformations), or renames the input file to '.orig' (on in-place transformations with the
 * 'keepOriginals' option).
 */
public
class RemoveFileTransformer implements FileTransformer {

    private static final Logger LOGGER = Logger.getLogger(RemoveFileTransformer.class.getName());

    private final boolean keepOriginals;

    public
    RemoveFileTransformer(boolean keepOriginals) {
        this.keepOriginals = keepOriginals;
    }

    @Override public void
    transform(String path, File in, File out) throws IOException {
        RemoveFileTransformer.LOGGER.log(Level.FINE, "Removing ''{0}''", in);

        if (in.equals(out)) {
            if (this.keepOriginals) {
                File origFile = new File(in.getParent(), '.' + in.getName() + ".orig");
                origFile.delete();
                if (!in.renameTo(origFile)) {
                    throw new IOException("Could not rename existing file '" + in + "' to '" + origFile + "'");
                }
            } else {
                if (!in.delete()) throw new IOException("Could not delete file '" + in + "'");
            }
        }
    }
}
