
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.filetransformation;

import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.unkrig.commons.file.FileUtil;
import de.unkrig.commons.text.pattern.Glob;

/**
 * Transforms files through a delegate {@link FileTransformer}, but modifies the output {@link File} iff the path of
 * the input {@link File} matches a given {@link Glob}.
 */
public
class RenameFileTransformer implements FileTransformer {

    private static final Logger LOGGER = Logger.getLogger(RenameFileTransformer.class.getName());

    private final Glob            glob;
    private final FileTransformer delegate;
    private final boolean         keepOriginals;

    public
    RenameFileTransformer(Glob glob, FileTransformer delegate, boolean keepOriginals) {
        this.glob          = glob;
        this.delegate      = delegate;
        this.keepOriginals = keepOriginals;
    }

    @Override public void
    transform(String path, File in, File out) throws IOException {
        String newName = this.glob.replace(out.getPath());
        if (newName == null) {

            // No renaming requested.
            this.delegate.transform(path, in, out);
            return;
        }

        File newOut = new File(newName);
        if (newOut.equals(out)) {

            // Redundant renaming requested.
            this.delegate.transform(path, in, out);
            return;
        }

        RenameFileTransformer.LOGGER.log(Level.FINE, "Renaming ''{0}'' to ''{1}''", new Object[] { in, newOut });

        if (!in.equals(out)) {

            // Out-of-place transformation with non-redundant renaming.
            if (in.equals(newOut)) {
                throw new IOException("Input file '" + in + "' and output file '" + out + "' collide due to renaming");
            }

            this.delegate.transform(path, in, newOut);
            return;
        }

        if (!this.keepOriginals) {

            // In-place transformation with non-redundant renaming in non-keep-originals mode.
            this.delegate.transform(path, in, newOut);
            if (!in.delete()) {
                FileUtil.attemptToDeleteRecursively(newOut);
                throw new IOException("Could not delete file '" + in + "'");
            }
        }

        // In-place transformation with non-redundant renaming in keep-originals mode.
        File origFile = new File(in.getParent(), '.' + in.getName() + ".orig");
        if (!origFile.delete()) {
            throw new IOException("File '" + origFile + "' is in the way and cannot be deleted");
        }
        try {
            this.delegate.transform(path, in, newOut);
        } catch (IOException ioe) {
            FileUtil.attemptToDeleteRecursively(newOut);
            throw ioe;
        }
        if (!in.renameTo(origFile)) {
            FileUtil.attemptToDeleteRecursively(newOut);
            throw new IOException("Could not rename existing file '" + in + "' to '" + origFile + "'");
        }
    }
}
