
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.filetransformation;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import de.unkrig.commons.file.zipentrytransformation.ZipEntriesTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryTransformer;
import de.unkrig.commons.lang.ExceptionUtil;

/**
 * @see #transform(String, File, File)
 */
public
class ZipFileTransformer implements FileTransformer {

    private final FileTransformer       defaultFileTransformer;
    private final ZipEntriesTransformer zipEntriesTransformer;
    private final ZipEntryTransformer   zipEntryTransformer;

    /**
     * @see #transform(String, File, File)
     */
    public
    ZipFileTransformer(
        FileTransformer       defaultFileTransformer,
        ZipEntriesTransformer zipEntriesTransformer,
        ZipEntryTransformer   zipEntryTransformer
    ) {
        this.defaultFileTransformer = defaultFileTransformer;
        this.zipEntriesTransformer  = zipEntriesTransformer;
        this.zipEntryTransformer    = zipEntryTransformer;
    }

    /**
     * If {@code in} is not in ZIP file format, then pass {@code in} and {@code out} to the {@code
     * defaultFileTransformer}. Otherwise, if the path of {@code in}, augmented with a '!', does not match the {@code
     * zipFilePathPredicate}, do nothing. Otherwise, a ZIP file {@code out} is created, and each ZIP entry of {@code
     * in} is passed to the {@code zipEntriesTransformer}, which eventually calls the {@code zipEntryTransformer}.
     */
    @Override public void
    transform(String path, File in, File out) throws FileNotFoundException, IOException {

        if (in.equals(out)) throw new AssertionError("IN and OUT must not be identical: " + in);

        // Attempt to transform the input file as a ZIP file.

        // Using "new ZipFile(file)" to check whether the file is in ZIP format is terribly expensive. For
        // performance reasons, we check ourselves by looking at the leading magic number.
        boolean isZip;
        {
            InputStream is = new FileInputStream(in);
            try {
                byte[] buf = new byte[4];
                isZip = (
                    is.read(buf) == 4
                    && buf[3] == 0x04 && buf[2] == 0x03 && buf[1] == 0x4B && buf[0] == 0x50
                );
            } catch (IOException ioe) {
                throw ExceptionUtil.wrap("Reading the first few bytes from '" + in + "'", ioe);
            } catch (RuntimeException re) {
                throw ExceptionUtil.wrap("Reading the first few bytes from '" + in + "'", re);
            } finally {
                try { is.close(); } catch (Exception ioe) {}
            }
        }

        if (!isZip) {
            this.defaultFileTransformer.transform(path, in, out);
            return;
        }

        String inPrefix = in.getPath() + '!';

        ZipFile zipFile = new ZipFile(in);
        try {

            OutputStream os = new FileOutputStream(out);
            try {

                // Create output ZIP file.
                ZipOutputStream zos = new ZipOutputStream(os);
                try {

                    // Transform each input ZIP entry into the output file.
                    for (Enumeration<? extends ZipEntry> entries = zipFile.entries(); entries.hasMoreElements();) {
                        ZipEntry    entry = entries.nextElement();
                        InputStream is    = zipFile.getInputStream(entry);
                        try {
                            try {
                                this.zipEntriesTransformer.visitEntry(
                                    inPrefix,
                                    entry,
                                    is,
                                    this.zipEntryTransformer,
                                    zos
                                );
                            } catch (IOException ioe) {
                                throw ExceptionUtil.wrap("Processing entry '" + entry.getName() + "'", ioe);
                            } catch (RuntimeException re) {
                                throw ExceptionUtil.wrap("Processing entry '" + entry.getName() + "'", re);
                            }
                            is.close();
                        } finally {
                            try { is.close(); } catch (Exception e) {}
                        }
                    }

                    this.zipEntriesTransformer.visitEnd(inPrefix, zos);

                    // Close output ZIP file.
                    zos.close();
                } finally {
                    try { zos.close(); } catch (Exception e) {}
                }
            } finally {
                try { os.close(); } catch (Exception e) {}
            }
            zipFile.close();
        } finally {
            try { zipFile.close(); } catch (Exception e) {}
        }
    }
}
