
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.zipentrytransformation;

import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import de.unkrig.commons.text.pattern.Glob;

/**
 * A {@link ZipEntriesTransformer} that renames zip entries.
 */
public
class RenameZipEntryTransformer implements ZipEntryTransformer {

    private static final Logger LOGGER = Logger.getLogger(RenameZipEntryTransformer.class.getName());

    private final Glob                glob;
    private final ZipEntryTransformer delegate;

    public
    RenameZipEntryTransformer(Glob glob, ZipEntryTransformer delegate) {
        this.glob     = glob;
        this.delegate = delegate;
    }

    @Override public void
    transform(ZipEntry entry, String path, InputStream is, ZipOutputStream zos) throws IOException {
        String newPath = this.glob.replace(path);

        if (newPath == null) {
            this.delegate.transform(entry, path, is, zos);
            return;
        }

        ZipEntry newEntry;
        {
            int idx = newPath.lastIndexOf('!');
            if (idx == -1 || !path.startsWith(newPath.substring(0, idx + 1))) {
                throw new IOException(
                    "Cannot rename ZIP entry '"
                    + path
                    + "' across ZIP file boundaries ('"
                    + newPath
                    + "')"
                );
            }
            RenameZipEntryTransformer.LOGGER.log(
                Level.FINE,
                "Renaming ''{0}'' to ''{1}''",
                new Object[] { path, newPath }
            );
            newEntry = new ZipEntry(newPath.substring(idx + 1));
            newEntry.setComment(entry.getComment());
            newEntry.setExtra(entry.getExtra());
            newEntry.setTime(entry.getTime());
        }

        this.delegate.transform(newEntry, newPath, is, zos);
    }
}
