
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2014, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.text.pattern;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.nio.CharBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/** {@link Pattern}-related utility methods. */
public final
class PatternUtil {

    private static final Logger LOGGER = Logger.getLogger(PatternUtil.class.getName());

    private PatternUtil() {}

    /**
     * Reads text from {@code in} replaces all occurrences of {@code pattern} with {@code replacement}, and
     * writes the result to {@code out}.
     */
    public static void
    replaceAll(Reader in, Pattern pattern, String replacement, Writer out) throws IOException {
        PatternUtil.replaceAll(in, pattern, replacement, out, 8192);
    }

    /**
     * Reads text from {@code in} replaces all occurrences of {@code pattern} with {@code replacement}, and
     * writes the result to {@code out}.
     */
    public static void
    replaceAll(Reader in, Pattern pattern, String replacement, Writer out, int initialBufferCapacity)
    throws IOException {

        int        substitutionCount = 0;
        CharBuffer cb                = CharBuffer.allocate(initialBufferCapacity);

        NEXT_CHUNK:
        for (;;) {

            // Fill the buffer as far as possible (limited by the buffer capacity or by EOI).
            while (cb.hasRemaining() && in.read(cb) != -1);
            if (cb.position() == 0) break; // End-of-input.
            cb.flip();

            // Find the next match.
            Matcher m     = pattern.matcher(cb);
            boolean found = m.find();

            while (m.hitEnd()) {

                // We hit the end; read more data until we don't hit the end any more.
                if (cb.limit() < cb.capacity()) {

                    // There's room left in the CharBuffer; fill it.
                    cb.compact();
                    if (in.read(cb) == -1) {

                        // End-of-input.
                        cb.flip();
                        break;
                    }
                    while (cb.hasRemaining() && in.read(cb) != -1);
                    cb.flip();
                } else {
                    PatternUtil.LOGGER.finest("Increasing buffer size");
                    cb = CharBuffer.allocate(cb.capacity() * 2).append(cb);
                    if (in.read(cb) == -1) {

                        // End-of-input.
                        cb.flip();
                        m     = pattern.matcher(cb);
                        found = m.find();
                        break;
                    }
                    while (cb.hasRemaining() && in.read(cb) != -1);
                    cb.flip();
                }

                m     = pattern.matcher(cb);
                found = m.find();
            }

            if (!found) {
                out.append(cb);
                cb.clear();
                continue NEXT_CHUNK;
            }

            if (m.end() == 0) {

                // Start-of-input match.
                if (cb.limit() == 0) break;
                out.append(cb.get());
                cb.compact();
                continue NEXT_CHUNK;
            }

            if (PatternUtil.LOGGER.isLoggable(Level.FINEST)) {
                PatternUtil.LOGGER.log(
                    Level.FINEST,
                    "Replacing ''{0}'' with ''{1}''",
                    new Object[] { cb.subSequence(m.start(), m.end()), replacement }
                );
            }

            StringBuffer sb = new StringBuffer();
            m.appendReplacement(sb, replacement);
            substitutionCount++;
            out.append(sb);
            cb.position(m.end());
            cb.compact();

            // If the CharBuffer was enlarged (due to 'hitEnd()'), shrink it to its initial size (if possible).
            if (cb.capacity() > initialBufferCapacity && cb.position() <= initialBufferCapacity) {
                PatternUtil.LOGGER.finest("Restoring initial buffer size");
                cb.flip();
                cb = CharBuffer.allocate(initialBufferCapacity).append(cb);
            }
        }

        PatternUtil.LOGGER.log(
            Level.FINE,
            "Replaced {0} match(es) of ''{1}'' with ''{2}''",
            new Object[] { substitutionCount, pattern, replacement }
        );
    }
}
