
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.Resource;
import org.apache.tools.ant.types.ResourceCollection;
import org.apache.tools.ant.types.resources.FileResource;

import de.unkrig.commons.lang.protocol.Predicate;
import de.unkrig.commons.lang.protocol.PredicateUtil;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.commons.text.pattern.Glob;
import de.unkrig.commons.text.pattern.Pattern2;
import de.unkrig.loggifier.LoggifyingClassAdapter.ErrorHandler;

/**
 * An ANT task for loggification of Java class files.
 *
 * To make this task accessible from your ANT scripts, inport it as follows:
 * <pre>&lt;taskdef
 *     classpath="lib/de.unkrig.loggifier.jar"
 *     resource="de/unkrig/loggifier/ant.properties"
 * /></pre>
 * @see Loggifier
 */
public
class AntTask extends Task {

    private static final Logger LOGGER = Logger.getLogger(AntTask.class.getName());

    @Nullable private File    file;
    @Nullable private File    dir;
    @Nullable private File    tofile;
    @Nullable private File    todir;
    private Predicate<String> selector = PredicateUtil.always();
    private boolean           keepOriginals;
    private final RuleSet     ruleSet = new RuleSet(LoggifyingContentsTransformer.DEFAULT_LOG_LEVEL_CALCULATOR);

    private final List<ResourceCollection> resourceCollections = new ArrayList<ResourceCollection>();

    /** Setter for 'selector' attribute. */
    public void
    setSelector(String value) { this.selector = Glob.compile(value, Pattern2.WILDCARD | Glob.INCLUDES_EXCLUDES); }

    /** Setter for 'file' attribute. */
    public void
    setFile(File value) { this.file = value; }

    /** Setter for 'dir' attribute. */
    public void
    setDir(File value) { this.dir = value; }

    /** Setter for 'tofile' attribute. */
    public void
    setTofile(File value) { this.tofile = value; }

    /** Setter for 'todir' attribute. */
    public void
    setTodir(File value) { this.todir = value; }

    /** Setter for 'keepOriginals' attribute. */
    public void
    setKeepOriginals(boolean value) { this.keepOriginals = value; }

    /** Helper bean for an element that has text, but no attributes and subelements. */
    public static
    class ElementWithText {

        private String text = "";

        /** Adder for element text. */
        public void
        addText(String text) { this.text = text; }
    }

    /** Added for 'rule' subelements. */
    public void
    addConfiguredRule(ElementWithText element) {
        try {
            this.ruleSet.add(element.text);
        } catch (Exception e) {
            AntTask.LOGGER.log(Level.SEVERE, e.toString(), e);
            throw new BuildException(e);
        }
    }

    /** Added for resource collection subelements. */
    public void
    addConfigured(ResourceCollection value) { this.resourceCollections.add(value); }

    /**
     * ANT execute method.
     */
    @Override public void
    execute() throws BuildException {
        try {
            final int[]  errorCount   = new int[1];
            ErrorHandler errorHandler = new LoggifyingClassAdapter.ErrorHandler() {

                @Override public void
                handle(String message) {
                    System.err.println(message);
                    errorCount[0]++;
                }

                @Override public void
                handle(Throwable t) {
                    t.printStackTrace();
                    errorCount[0]++;
                }
            };

            Loggifier loggifier = new Loggifier(this.selector, this.ruleSet, this.keepOriginals, errorHandler);

            // Handle 'file' / 'tofile' / 'todir'.
            {
                File file   = this.file;
                File tofile = this.tofile;

                if (file != null) {
                    if (tofile == null) {
                        loggifier.transform(
                            file.getPath(),
                            file,
                            this.todir == null ? file : new File(this.todir, file.getName())
                        );
                    } else
                    if (this.todir == null) {
                        loggifier.transform(file.getPath(), file, tofile);
                    } else
                    {
                        throw new BuildException("'tofile=\"...\"' and 'todir=\"...\"' are mutually exclusive");
                    }
                } else
                if (tofile != null) {
                    throw new BuildException(
                        "'tofile=\"...\"' must only be configured in connection with 'file=\"...\"'"
                    );
                }
            }

            // Handle 'dir' / 'todir'
            {
                File dir   = this.dir;
                File todir = this.todir;

                if (dir != null) {
                    loggifier.transform(dir.getPath(), dir, todir != null ? todir : dir);
                }
            }

            // Handle resource collections.
            for (ResourceCollection resourceCollection : this.resourceCollections) {
                for (
                    @SuppressWarnings("unchecked") Iterator<Resource> it = resourceCollection.iterator();
                    it.hasNext();
                ) {
                    Resource resource = it.next();
                    if (resource.isFilesystemOnly()) {
                        FileResource fileResource = (FileResource) resource;
                        File         file         = fileResource.getFile();
                        loggifier.transform(
                            fileResource.getName(),
                            file,
                            this.todir == null ? file : new File(this.todir, fileResource.getName())
                        );
                    } else {
                        throw new UnsupportedOperationException();
                    }
                }
            }

            if (errorCount[0] > 0) throw new BuildException(errorCount[0] + " error(s) while loggifying");
        } catch (Exception e) {
            AntTask.LOGGER.log(Level.SEVERE, e.toString(), e);
            throw new BuildException(e);
        }
    }
}
