
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import java.io.ByteArrayInputStream;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;

import de.unkrig.commons.nullanalysis.NotNullByDefault;

/**
 * This class plugs into the
 * <a href="http://docs.oracle.com/javase/7/docs/api/java/lang/instrument/package-summary.html">"Java programming
 * language agents" API</a>
 * (available since Java 5). It implements loggification at class-loading time.
 * <p>
 * To enable loggification at class-loading time, add
 * <pre>-javaagent:de.unkrig.loggifier.jar</pre>
 * to the JAVA command line.
 */
@NotNullByDefault(false) public // SUPPRESS CHECKSTYLE HideUtilityClassConstructor
class InstrumentationAgent {

    /**
     * @see <a href="http://docs.oracle.com/javase/7/docs/api/java/lang/instrument/package-summary.html">Description
     *      of package java.lang.instrument</a>
     */
    public static void
    premain(String agentArgs, Instrumentation instrumentation) {

        // Compile the 'agentArgs' as loggification rules.
        RuleSet ruleSet = new RuleSet(LoggifyingContentsTransformer.DEFAULT_LOG_LEVEL_CALCULATOR);
        if (agentArgs != null) {
            try {
                for (String agentArg : agentArgs.split("[\r\n]+")) {
                    ruleSet.add(agentArg);
                }
            } catch (Exception e) {
                System.err.println("Processing argument of '-javaagent:de.unkrig.loggifier.jar': " + e.getMessage());
            }
        }

        final LoggifyingContentsTransformer loggifyingContentsTransformer = new LoggifyingContentsTransformer(
            ruleSet,
            LoggifyingClassAdapter.ErrorHandler.IGNORE
        );

        instrumentation.addTransformer(new ClassFileTransformer() {

            @Override public byte[]
            transform(
                ClassLoader      classLoader,
                String           className,
                Class<?>         classBeingRedefined,
                ProtectionDomain protectionDomain,
                byte[]           data
            ) {
                if (classLoader != ClassLoader.getSystemClassLoader()) {
                    return data;
                }

                if (className.startsWith("de/unkrig/loggifier/runtime/")) {
                    return data;
                }

                try {
                    return loggifyingContentsTransformer.transform(className, new ByteArrayInputStream(data));
                } catch (Exception e) {
                    e.printStackTrace();
                    return data;
                }
            }
        });
    }
}
