
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.ByteVector;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.signature.SignatureReader;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.IincInsnNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.InsnNode;
import org.objectweb.asm.tree.IntInsnNode;
import org.objectweb.asm.tree.JumpInsnNode;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.LineNumberNode;
import org.objectweb.asm.tree.LocalVariableNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.MultiANewArrayInsnNode;
import org.objectweb.asm.tree.TryCatchBlockNode;
import org.objectweb.asm.tree.TypeInsnNode;
import org.objectweb.asm.tree.VarInsnNode;
import org.objectweb.asm.util.TraceSignatureVisitor;

import de.unkrig.commons.asm.InsnUtil;
import de.unkrig.commons.asm.OpcodeUtil;
import de.unkrig.commons.asm.OpcodeUtil.InsnDescription;
import de.unkrig.commons.asm.Signature;
import de.unkrig.commons.asm.Types;
import de.unkrig.commons.lang.ExceptionUtil;
import de.unkrig.commons.lang.PrettyPrinter;
import de.unkrig.commons.nullanalysis.NotNull;
import de.unkrig.commons.nullanalysis.NotNullByDefault;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.commons.text.parser.ParseException;
import de.unkrig.loggifier.LoggifyingContentsTransformer.Action;
import de.unkrig.loggifier.loglevel.LevelCalculator;
import de.unkrig.loggifier.loglevel.LevelCalculatorUtil;
import de.unkrig.loggifier.runtime.Helper;
import de.unkrig.loggifier.runtime.annotation.Loggify;

/**
 * Acceptors may throw {@link ExceptionTableEntryOffsetTooLargeException}s iff a try-catch block grows too large during
 * loggification.
 * <p>
 * Acceptors may throw {@link DoubleLoggificationException}s iff the subject class is already loggified.
 */
public
class LoggifyingClassAdapter extends ClassVisitor implements Opcodes {

    private static final String STATISTICS_ATTRIBUTE_TYPE = "de.unkrig.loggifier.Statistics";

    /** The internal name of {@link Helper}. */
    static final String HELPER_CLASS_NAME = "de/unkrig/loggifier/runtime/Helper";

    /**
     * The format for the logger names at CLASS scope.
     * <p>
     * {@link MessageFormat} placeholders are:
     * <dl>
     *   <dt>{0}<dd>Package prefix,                     e.g. '{@code de.unkrig.loggifier.}'
     *   <dt>{1}<dd>Unqualified top-level type name,    e.g. '{@code OuterClass}'
     *   <dt>{2}<dd>Nested type suffix,                 e.g. '{@code $NestedClass$NestedNestedClass}'
     *   <dt>{6}<dd>Action name,                        e.g. '{@code ENTRY}'
     * </dl>
     */
    static final MessageFormat CLASS_LOGGER_NAME_FORMAT  = new MessageFormat("{0}{1}{2}.{6}");

    /**
     * The format for the logger names at METHOD scope.
     * <p>
     * {@link MessageFormat} placeholders are:
     * <dl>
     *   <dt>{0}<dd>Package prefix,                     e.g. '{@code de.unkrig.loggifier.}'
     *   <dt>{1}<dd>Unqualified top-level type name,    e.g. '{@code OuterClass}'
     *   <dt>{2}<dd>Nested type suffix,                 e.g. '{@code $NestedClass$NestedNestedClass}'
     *   <dt>{3}<dd>(Dequalified) return type name,     e.g. '{@code String}'
     *   <dt>{4}<dd>Method name,                        e.g. '{@code myMethod}'
     *   <dt>{5}<dd>(Dequalified) parameter type names, e.g. '{@code ()}' or '{@code (int, String)}', or '{@code
     *                                                                                     <Class<T>>(List<String)}'
     *   <dt>{6}<dd>Action name,                        e.g. '{@code ENTRY}'
     * </dl>
     */
    static final MessageFormat METHOD_LOGGER_NAME_FORMAT = new MessageFormat("{0}{1}{2}.{4}{5}.{6}");

    /** The name of the magic field through which additional rules can be configured. */
    static final Object RULE_FIELD_NAME = "$$loggifierRules";

    /** The descriptor of the {@link Loggify} annotation. */
    private static final String LOGGIFY_ANNOTATION_DESCRIPTOR = "Lde/unkrig/loggifier/runtime/annotation/Loggify;";

    /** An empty {@link InsnList}. */
    public static final InsnList NO_INSNS = new InsnList();

    /** For {@link AbstractInsnNode#clone(Map)}. */
    public static final Map<LabelNode, LabelNode> IDENTITY;
    static {
        @NotNullByDefault(false)
        class IdentityMap implements Map<LabelNode, LabelNode> {

            @Override public int
            size() { throw new UnsupportedOperationException("size"); }

            @Override public boolean
            isEmpty() { return false; }

            @Override public boolean
            containsKey(Object key) { return true; }

            @Override public boolean
            containsValue(Object value) { return true; }

            @Override public LabelNode
            get(Object key) { return (LabelNode) key; }

            @Override public LabelNode
            put(LabelNode key, LabelNode value) { throw new UnsupportedOperationException("put"); }

            @Override public LabelNode
            remove(Object key) { throw new UnsupportedOperationException("remove"); }

            @Override public void
            putAll(Map<? extends LabelNode, ? extends LabelNode> m) {
                throw new UnsupportedOperationException("putAll");
            }

            @Override public void
            clear() { throw new UnsupportedOperationException("clear"); }

            @Override public Set<LabelNode>
            keySet() { throw new UnsupportedOperationException("keySet"); }

            @Override public Collection<LabelNode>
            values() { throw new UnsupportedOperationException("size"); }

            @Override public Set<java.util.Map.Entry<LabelNode, LabelNode>>
            entrySet() { throw new UnsupportedOperationException("values"); }
        }
        IDENTITY = new IdentityMap();
    }

    private LevelCalculator    classLogLevelCalculator;
    private final ErrorHandler errorHandler;

    /**
     * @see Opcodes
     */
    private int thisClassAccess;

    /**
     * Internal name of the class being visited, e.g. '{@code pkg/MyClass}'.
     */
    @Nullable private String thisClassName;

    /**
     * Fully qualified name of the class being visited, e.g. '{@code pkg.MyClass}'.
     */
    @Nullable private String sourceClassName;

    @Nullable private String        source;
    @Nullable private MethodNode    classInitializer;
    private final SortedSet<String> loggerNames = new TreeSet<String>();

    /**
     * How many loggings were injected into this CLASS file.
     */
    private int loggingCount;

    /**
     * @param logLevelCalculators This map determines for which code actions logging code is injected, and on what
     *                            level the action is logged. <b>Must</b> contain entries for <b>all</b> values of
     *                            {@link Action}.
     */
    public
    LoggifyingClassAdapter(ClassVisitor cv, LevelCalculator logLevelCalculator, ErrorHandler errorHandler) {
        super(Opcodes.ASM4, cv);
        this.classLogLevelCalculator = logLevelCalculator;
        this.errorHandler            = errorHandler;
    }

    @Override public void
    visit(
        int                              version,
        int                              classAccess,
        @SuppressWarnings("null") String thisClassName,
        @Nullable String                 optionalClassSignature,
        @Nullable String                 optionalSuperclassName,
        @Nullable String[]               optionalImplementedInterfaceNames
    ) {
        this.thisClassAccess = classAccess;
        this.thisClassName   = thisClassName;
        this.sourceClassName = thisClassName.replace('/', '.');

        this.source           = null;
        this.classInitializer = null;
        this.loggerNames.clear();

        super.visit(
            version,
            classAccess,
            thisClassName,
            optionalClassSignature,
            optionalSuperclassName,
            optionalImplementedInterfaceNames
        );

        // Add a custom attribute documenting the loggification.
        this.cv.visitAttribute(new Attribute(LoggifyingClassAdapter.STATISTICS_ATTRIBUTE_TYPE) {

            @Override protected ByteVector
            write(
                @Nullable ClassWriter cw,
                @Nullable byte[]      code,
                int                   len,
                int                   maxStack,
                int                   maxLocals
            ) {
                return new ByteVector().putUTF8(
                    "Inserted "
                    + LoggifyingClassAdapter.this.loggingCount
                    + " logging points"
                );
            }
        });
    }

    @Override @NotNullByDefault(false) public void
    visitAttribute(Attribute attribute) {
        if (attribute.type.equals(LoggifyingClassAdapter.STATISTICS_ATTRIBUTE_TYPE)) {
            throw new DoubleLoggificationException("Class is already loggified");
        }
        super.visitAttribute(attribute);
    }

    @Override @NotNullByDefault(false) public void
    visitSource(String source, String debug) {
        this.source = source;
        super.visitSource(source, debug);
    }

    @Override @NotNullByDefault(false) public AnnotationVisitor
    visitAnnotation(final String annotationDescriptor, final boolean visible) {
        AnnotationVisitor delegate = super.visitAnnotation(annotationDescriptor, visible);
        if (!annotationDescriptor.equals(LoggifyingClassAdapter.LOGGIFY_ANNOTATION_DESCRIPTOR)) {
            return delegate;
        }

        return new AnnotationVisitor(Opcodes.ASM4, delegate) {

            @Override public AnnotationVisitor
            visitArray(String name) {
                return new AnnotationVisitor(Opcodes.ASM4, super.visitArray(name)) {

                    private final List<String> lines = new ArrayList<String>();

                    @Override public void
                    visit(String name, Object value) {
                        this.lines.add((String) value);
                        super.visit(name, value);
                    }

                    @Override @NotNullByDefault(false) public void
                    visitEnd() {

                        // Parse the values of the @Loggify type annotation.
                        LoggifyingClassAdapter.this.classLogLevelCalculator = (
                            LoggifyingClassAdapter.parseLoggifyAnnotation(
                                this.lines.toArray(new String[this.lines.size()]),
                                LoggifyingClassAdapter.this.classLogLevelCalculator,
                                new ErrorHandler() {

                                    @Override @NotNullByDefault public void
                                    handle(Throwable t) {
                                        LoggifyingClassAdapter.this.errorHandler.handle(ExceptionUtil.wrap((
                                            "Parsing or evaluating the @Loggify annotation of class '"
                                            + LoggifyingClassAdapter.this.sourceClassName
                                            + "'"
                                        ), t));
                                    }

                                    @Override @NotNullByDefault public void
                                    handle(String message) {
                                        LoggifyingClassAdapter.this.errorHandler.handle(
                                            "Parsing or evaluating the @Loggify annotation of class '"
                                            + LoggifyingClassAdapter.this.sourceClassName
                                            + "': "
                                            + message
                                        );
                                    }
                                }
                            )
                        );

                        super.visitEnd();
                    }
                };
            }
        };
    }

    @Override public void
    visitEnd() {

        LOGGIFY_CLINIT: {

            String sourceClassName = this.sourceClassName;
            assert sourceClassName != null;
            Level logLevel = this.classLogLevelCalculator.calculate(
                Action.CLINIT,                           // action
                this.source,                             // source
                this.thisClassAccess,                    // classAccess
                sourceClassName,                         // className
                Opcodes.ACC_STATIC | Opcodes.ACC_PUBLIC, // methodAccess
                "<clinit>()",                            // methodName
                1,                                       // lineNumber
                null                                     // custom
            );
            if (logLevel == Level.OFF) break LOGGIFY_CLINIT;

            // Create a class initializer if none exists.
            MethodNode ci = this.classInitializer;
            if (ci == null) {
                ci = (
                    this.classInitializer
                    = new MethodNode(Opcodes.ACC_PUBLIC | Opcodes.ACC_STATIC, "<clinit>", "()V", null, null)
                );
                ci.instructions.add(new InsnNode(Opcodes.RETURN));
            }

            InsnList insns = ci.instructions;
            insns.insertBefore(
                insns.getFirst(),
                this.generateLogging(
                    this.getLoggerName(Action.CLINIT),              // loggerName
                    logLevel,                                       // logLevel
                    "<clinit>()",                                   // sourceMethodName
                    1,                                              // line
                    "CLINIT " + sourceClassName,                    // message
                    InsnUtil.il(new InsnNode(Opcodes.ACONST_NULL)), // pushParameters
                    ci                                              // methodVisitor
                )
            );
        }

        // Add static logger fields:
        //    static Logger <logger-field-name>;
        for (String loggerName : this.loggerNames) {
            new FieldNode(
                (
                    (this.thisClassAccess & Opcodes.ACC_INTERFACE) != 0
                    ? Opcodes.ACC_PUBLIC | Opcodes.ACC_STATIC | Opcodes.ACC_FINAL
                    : Opcodes.ACC_PRIVATE | Opcodes.ACC_STATIC | Opcodes.ACC_FINAL
                ),
                LoggifyingClassAdapter.loggerNameToFieldName(loggerName),
                "Ljava/util/logging/Logger;",
                null,
                null
            ).accept(this);
        }

        if (!this.loggerNames.isEmpty()) {
            MethodNode ci = this.classInitializer;
            if (ci == null) {
                ci = (
                    this.classInitializer
                    = new MethodNode(Opcodes.ACC_PUBLIC | Opcodes.ACC_STATIC, "<clinit>", "()V", null, null)
                );
                ci.instructions.add(new InsnNode(Opcodes.RETURN));
            }

            // Inject logger creation code.
            InsnList il                 = new InsnList();
            String   previousLoggerName = "";
            for (String loggerName : this.loggerNames) {

                // Create intermediate loggers. E.g. if there are three loggers
                //     a.b
                //     a.b.c.d
                //     a.b.c.e
                // , then we want to have an intermediate logger "a.b.c", so the user can easily configure the loggers
                // "a.b.c.d" and "a.b.c.e" together.
                {
                    String commonPrefix = loggerName.substring(0, loggerName.lastIndexOf('.') + 1);
                    while (!previousLoggerName.startsWith(commonPrefix)) {
                        commonPrefix = commonPrefix.substring(
                            0,
                            commonPrefix.lastIndexOf('.', commonPrefix.length() - 2) + 1
                        );
                    }
                    for (
                        int idx = loggerName.indexOf('.', commonPrefix.length());
                        idx != -1;
                        idx = loggerName.indexOf('.', idx + 1)
                    ) {
                        String intermediateLoggerName = loggerName.substring(0, idx);

                        // Logger.getLogger("<intermediate-logger-name>");
                        il.add(new LdcInsnNode(intermediateLoggerName));
                        il.add(new MethodInsnNode(
                            Opcodes.INVOKESTATIC,
                            "java/util/logging/Logger",
                            "getLogger",
                            "(Ljava/lang/String;)Ljava/util/logging/Logger;"
                        ));
                        il.add(new InsnNode(Opcodes.POP));
                    }
                    previousLoggerName = loggerName;
                }

                // <this-class-name>.<logger-field-name> = Logger.getLogger("<logger-name>");
                il.add(new LdcInsnNode(loggerName));
                il.add(new MethodInsnNode(
                    Opcodes.INVOKESTATIC,
                    "java/util/logging/Logger",
                    "getLogger",
                    "(Ljava/lang/String;)Ljava/util/logging/Logger;"
                ));
                il.add(new FieldInsnNode(
                    Opcodes.PUTSTATIC,
                    this.thisClassName,
                    LoggifyingClassAdapter.loggerNameToFieldName(loggerName),
                    "Ljava/util/logging/Logger;"
                ));
            }

            InsnList insns = ci.instructions;
            insns.insertBefore(insns.getFirst(), il);
        }

        MethodNode ci = this.classInitializer;
        if (ci != null) {
            ci.accept(this.cv);
            LoggifyingClassAdapter.checkTryCatchBlocks(ci);
        }

        super.visitEnd();
    }

    /**
     * As of version 3.3.1, ASM does not verify that the offsets in a {@link TryCatchBlockNode} are 0xffff or less.
     * Since the consequences are always fatal, we need to check for ourselves.
     *
     * @throw ExceptionTableEntryOffsetTooLargeException
     */
    public static void
    checkTryCatchBlocks(MethodNode methodNode) {

        @SuppressWarnings("unchecked") List<TryCatchBlockNode> tcbs = methodNode.tryCatchBlocks;
        for (TryCatchBlockNode tcbn : tcbs) {
            if (
                tcbn.start.getLabel().getOffset() > 0xffff
                || tcbn.end.getLabel().getOffset() > 0xffff
                || tcbn.handler.getLabel().getOffset() > 0xffff
            ) {
                throw new ExceptionTableEntryOffsetTooLargeException(
                    "Bytecode of method '"
                    + Signature.toString(methodNode)
                    + "grew too large during loggification - please tighten the loggification rules for this method"
                );
            }
        }
    }

    /**
     * Each logger needs a static field, and the name of that static field is derived from the logger name.
     */
    private static String
    loggerNameToFieldName(String loggerName) {
        StringBuilder sb = new StringBuilder("__");
        for (char c : loggerName.toCharArray()) {
            switch (c) {
            case '_':
                sb.append("__");
                break;
            case '.':
                sb.append("_dot_");
                break;
            case '[':
                sb.append("_lbracket_");
                break;
            case ']':
                sb.append("_rbracket_");
                break;
            case '(':
                sb.append("_lparen_");
                break;
            case ')':
                sb.append("_rparen_");
                break;
            case '<':
                sb.append("_lt_");
                break;
            case '>':
                sb.append("_gt_");
                break;
            default:
                if (Character.isJavaIdentifierPart(c)) {
                    sb.append(c);
                } else {
                    sb.append('_').append((int) c).append('_');
                }
            }
        }

        return sb.toString();
    }

    @Override @NotNullByDefault(false) public FieldVisitor
    visitField(
        final int    access,
        final String name,
        final String desc,
        final String signature,
        final Object value
    ) {
        if (name.equals(LoggifyingClassAdapter.RULE_FIELD_NAME) && value instanceof String) {
            LoggifyingClassAdapter.this.classLogLevelCalculator = LoggifyingClassAdapter.parseLoggifyAnnotation(
                ((String) value).split("\n"),
                LoggifyingClassAdapter.this.classLogLevelCalculator,
                new ErrorHandler() {

                    @Override @NotNullByDefault public void
                    handle(Throwable t) {
                        LoggifyingClassAdapter.this.errorHandler.handle(ExceptionUtil.wrap((
                            "Parsing the '"
                            + LoggifyingClassAdapter.RULE_FIELD_NAME
                            + "' field of class '"
                            + LoggifyingClassAdapter.this.sourceClassName
                            + "'"
                        ), t));
                    }

                    @Override @NotNullByDefault public void
                    handle(String message) {
                        LoggifyingClassAdapter.this.errorHandler.handle(
                            "Parsing the '"
                            + LoggifyingClassAdapter.RULE_FIELD_NAME
                            + "' field of class '"
                            + LoggifyingClassAdapter.this.sourceClassName
                            + "': "
                            + message
                        );
                    }
                }
            );
        }

        return super.visitField(access, name, desc, signature, value);
    }

    @Override @NotNullByDefault(false) public MethodVisitor
    visitMethod(
        final int      methodAccess,
        final String   methodName,
        final String   methodDescriptor,
        final String   optionalMethodSignature,
        final String[] methodExceptions
    ) {
        final
        class LoggifyingMethodVisitor extends MethodNode {

            LevelCalculator methodLogLevelCalculator = LoggifyingClassAdapter.this.classLogLevelCalculator;

            /**
             * The (enhanced) name of method currently being visisted, e.g.
             * <pre>
             * meth(String, MyClass)
             * &lt;T extends IOException>meth(T, MyClass, java.lang.List&lt;T>)
             * </pre>
             * Notice that all type names are 'de-qualified', i.e. only their simple names appear.
             */
            private final String sourceMethodName = LoggifyingClassAdapter.dequalify(
                Signature.toString(methodName, methodDescriptor, optionalMethodSignature)
            );

            private LocalVariableFinder localVariableFinder;
            private CurrentLineProvider currentLineProvider;

            private
            LoggifyingMethodVisitor() {
                super(methodAccess, methodName, methodDescriptor, optionalMethodSignature, methodExceptions);
            }

            @Override public AnnotationVisitor
            visitAnnotation(final String annotationDescriptor, final boolean visible) {
                AnnotationVisitor delegate = super.visitAnnotation(annotationDescriptor, visible);
                if (!annotationDescriptor.equals(LoggifyingClassAdapter.LOGGIFY_ANNOTATION_DESCRIPTOR)) {
                    return delegate;
                }

                return new AnnotationVisitor(Opcodes.ASM4, delegate) {

                    @Override public AnnotationVisitor
                    visitArray(String name) {
                        return new AnnotationVisitor(Opcodes.ASM4, super.visitArray(name)) {

                            private final List<String> lines = new ArrayList<String>();

                            @Override public void
                            visit(String name, Object value) {
                                this.lines.add((String) value);
                                super.visit(name, value);
                            }

                            @Override @NotNullByDefault(false) public void
                            visitEnd() {

                                // Parse the values of the @Loggify type annotation.
                                LoggifyingMethodVisitor.this.methodLogLevelCalculator = (
                                    LoggifyingClassAdapter.parseLoggifyAnnotation(
                                        this.lines.toArray(new String[this.lines.size()]),
                                        LoggifyingMethodVisitor.this.methodLogLevelCalculator,
                                        new ErrorHandler() {

                                            @Override @NotNullByDefault public void
                                            handle(Throwable t) {
                                                LoggifyingClassAdapter.this.errorHandler.handle(ExceptionUtil.wrap((
                                                    "Parsing the @Loggify annotation of method '"
                                                    + methodName
                                                    + "' of class '"
                                                    + LoggifyingClassAdapter.this.sourceClassName
                                                    + "'"
                                                ), t));
                                            }

                                            @Override @NotNullByDefault public void
                                            handle(String message) {
                                                LoggifyingClassAdapter.this.errorHandler.handle(
                                                    "Parsing the @Loggify annotation of method '"
                                                    + methodName
                                                    + "' of class '"
                                                    + LoggifyingClassAdapter.this.sourceClassName
                                                    + "': "
                                                    + message
                                                );
                                            }
                                        }
                                    )
                                );

                                super.visitEnd();
                            }
                        };
                    }
                };
            }

            /**
             * @throw ExceptionTableEntryOffsetTooLargeException
             */
            @Override public void
            visitEnd() {
                try {
                    this.cook();
                    if ("<clinit>".equals(methodName)) {
                        LoggifyingClassAdapter.this.classInitializer = this;
                    } else {

                        @SuppressWarnings("synthetic-access") ClassVisitor cv = LoggifyingClassAdapter.this.cv;
                        this.accept(cv);
                        LoggifyingClassAdapter.checkTryCatchBlocks(this);
                    }
                } catch (RuntimeException re) {
                    throw ExceptionUtil.wrap((
                        "Loggifying class '"
                        + LoggifyingClassAdapter.this.sourceClassName
                        + "', method '"
                        + this.sourceMethodName
                        + "'"
                    ), re);
                }
            }

            private void
            cook() {
                if (methodName.startsWith("access$")) return;

                final Map<LabelNode, Collection<LocalVariableNode>> lvstart, lvend;
                if (this.localVariables == null) {
                    lvstart = (lvend = Collections.emptyMap());
                } else {
                    lvstart = new HashMap<LabelNode, Collection<LocalVariableNode>>();
                    lvend   = new HashMap<LabelNode, Collection<LocalVariableNode>>();

                    @SuppressWarnings("unchecked") List<LocalVariableNode> lvs = this.localVariables;
                    for (LocalVariableNode lvn : lvs) {
                        {
                            Collection<LocalVariableNode> c = lvstart.get(lvn.start);
                            if (c == null) lvstart.put(lvn.start, (c = new ArrayList<LocalVariableNode>()));
                            c.add(lvn);
                        }

                        {
                            Collection<LocalVariableNode> c = lvend.get(lvn.end);
                            if (c == null) lvend.put(lvn.end, (c = new ArrayList<LocalVariableNode>()));
                            c.add(lvn);
                        }
                    }

                }

                class LocalVariableTrackingInsnIterator
                implements Iterator<AbstractInsnNode>, LocalVariableFinder, CurrentLineProvider {

                    private final ListIterator<AbstractInsnNode> delegate;
                    {
                        @SuppressWarnings("unchecked") ListIterator<AbstractInsnNode> tmp = (
                            LoggifyingMethodVisitor.this.instructions.iterator()
                        );
                        this.delegate = tmp;
                    }

                    private final Map<Integer, LocalVariableNode> effectiveLocalVariables = (
                        new HashMap<Integer, LocalVariableNode>()
                    );
                    private int currentLine;

                    @Override public boolean
                    hasNext() {
                        return this.delegate.hasNext();
                    }

                    @Override public AbstractInsnNode
                    next() {
                        AbstractInsnNode result = this.delegate.next();
                        switch (result.getType()) {

                        case AbstractInsnNode.LABEL:
                            LabelNode ln = (LabelNode) result;

                            {
                                Collection<LocalVariableNode> c = lvstart.get(ln);
                                if (c != null) {
                                    for (LocalVariableNode lvn : c) {
                                        LocalVariableNode prev = this.effectiveLocalVariables.put(
                                            Integer.valueOf(lvn.index),
                                            lvn
                                        );
                                        if (prev != null) {
                                            throw new IllegalStateException(
                                                "More than one LV effctive with index "
                                                + lvn.index
                                                + ": '"
                                                + prev.name
                                                + "' and '"
                                                + lvn.name
                                                + "'"
                                            );
                                        }
                                    }
                                }
                            }

                            {
                                Collection<LocalVariableNode> c = lvend.get(ln);
                                if (c != null) {
                                    for (LocalVariableNode lvn : c) {
                                        if (this.effectiveLocalVariables.remove(Integer.valueOf(lvn.index)) != lvn) {
                                            throw new IllegalStateException(
                                                "Inconsistent LV visibility for index "
                                                + lvn.index
                                            );
                                        }
                                    }
                                }
                            }
                            break;

                        case AbstractInsnNode.LINE:
                            LineNumberNode lnn = (LineNumberNode) result;
                            this.currentLine = lnn.line;
                            break;

                        default:
                            ;
                        }
                        return result;
                    }

                    @Override public void
                    remove() {
                        this.delegate.remove();
                    }

                    @Override public LocalVariableNode
                    getLocalVariable(int index) {
                        return this.effectiveLocalVariables.get(Integer.valueOf(index));
                    }

                    @Override public int
                    getCurrentLine() {
                        return this.currentLine;
                    }
                }
                LocalVariableTrackingInsnIterator it = new LocalVariableTrackingInsnIterator();
                if (!(it.hasNext())) return;
                this.localVariableFinder = it;
                this.currentLineProvider = it;
                AbstractInsnNode insn = it.next(); // One instruction look-ahead.

                // If this is a constructor, skip all alternate constructor invocations.
                if ("<init>".equals(methodName)) {
                    int pendingNews = 0;
                    for (;;) {
                        int opcode = insn.getOpcode();

                        // The ECLIPSE  Java compiler sometimes generates
                        //     throw new java.lang.Error("Unresolved compilation problem");
                        // instead of a proper constructor (which starts with a superconstructor invocation).
                        if (opcode == Opcodes.ATHROW) return;

                        insn = it.next();
                        if (opcode == Opcodes.INVOKESPECIAL) {
                            if (pendingNews == 0) break;
                            --pendingNews;
                        } else
                        if (opcode == Opcodes.NEW) {
                            ++pendingNews;
                        }
                    }
                }

                // Skip all non-instructions (labels etc.).
                while (insn.getOpcode() == -1) {
                    insn = it.next();
                }

                // Inject method entry logging code.
                LOGGIFY_ENTRY: {
                    Level logLevel = this.getLogLevel(Action.ENTRY, null);
                    if (logLevel == Level.OFF) break LOGGIFY_ENTRY;

                    StringBuilder sb = new StringBuilder("ENTRY ");

                    Type[] parameterTypes = Type.getArgumentTypes(methodDescriptor);

                    Type[] localVariableTypes;
                    int    idx = 0;
                    if ((methodAccess & Opcodes.ACC_STATIC) == 0) {
                        sb.append(" {" + idx++ + "}");
                        localVariableTypes    = new Type[parameterTypes.length + 1];
                        localVariableTypes[0] = Types.OBJECT_TYPE;
                        System.arraycopy(parameterTypes, 0, localVariableTypes, 1, parameterTypes.length);
                    } else {
                        localVariableTypes = parameterTypes;
                    }

                    sb.append(" args=(");
                    if (parameterTypes.length > 0) {
                        sb.append("{" + idx++ + "}");
                        for (int i = parameterTypes.length - 1; i > 0; i--) {
                            sb.append(", {" + idx++ + "}");
                        }
                    }
                    sb.append(")");

                    this.instructions.insertBefore(insn, this.generateLoggingWithLocalVariables(
                        this.getLoggerName(Action.ENTRY), // loggerName
                        logLevel,                         // logLevel
                        sb.toString(),                    // message
                        0,                                // firstLocalVariableIndex
                        localVariableTypes,               // localVariableTypes
                        true                              // withNames
                    ));
                }

                // Now loggify the rest of the method's code.
                while (it.hasNext()) {

                    // Determine the next insn BEFORE calling 'loggifyInsn()', because 'loggifyInsn()' may insert
                    // code after 'insn'.
                    AbstractInsnNode next = it.next();
                    this.loggifyInsn(insn);
                    insn = next;
                }
                this.loggifyInsn(insn);

                // Now examine the method's exception table and inject CATCH logging for all exception handlers.
                @SuppressWarnings("unchecked") List<TryCatchBlockNode> tmp = this.tryCatchBlocks;
                for (TryCatchBlockNode tcbn : tmp) {
                    int tmpVar = this.maxLocals;
                    this.instructions.insert(tcbn.handler, LoggifyingClassAdapter.this.generateLoggingWithThrowable(
                        this.getLoggerName(Action.CATCH),                     // loggerName
                        this.getLogLevel(Action.CATCH, null),                 // logLevel
                        InsnUtil.il(new VarInsnNode(Opcodes.ASTORE, tmpVar)), // optionalPrefix
                        this.sourceMethodName,                                // sourceMethodName
                        this.currentLineProvider.getCurrentLine(),            // line
                        "CATCH  " + (                                         // message
                            tcbn.type == null ? "any" : tcbn.type.replace('/', '.')
                        ),
                        InsnUtil.il(new VarInsnNode(Opcodes.ALOAD, tmpVar)),  // pushThrowable
                        InsnUtil.il(new VarInsnNode(Opcodes.ALOAD, tmpVar)),  // optionalSuffix
                        this                                                  // methodVisitor
                    ));
                }
            }

            private void
            loggifyInsn(final AbstractInsnNode insn) {
                int opcode = insn.getOpcode();

                switch (opcode) {

                case ACONST_NULL:
                    this.loggifyConst(insn, "CONST  null");
                    break;
                case ICONST_M1:
                case ICONST_0: case ICONST_1: case ICONST_2: case ICONST_3: case ICONST_4: case ICONST_5:
                    this.loggifyConst(insn, "CONST  " + (opcode - Opcodes.ICONST_0));
                    break;
                case LCONST_0: case LCONST_1:
                    this.loggifyConst(insn, "CONST  " + (opcode - Opcodes.LCONST_0) + 'L');
                    break;
                case FCONST_0: case FCONST_1: case FCONST_2:
                    this.loggifyConst(insn, "CONST  " + (opcode - Opcodes.FCONST_0) + 'F');
                    break;
                case DCONST_0: case DCONST_1:
                    this.loggifyConst(insn, "CONST  " + (opcode - Opcodes.DCONST_0) + 'D');
                    break;
                case BIPUSH: case SIPUSH:
                    this.loggifyConst(insn, "CONST  " + ((IntInsnNode) insn).operand);
                    break;
                case LDC:
                    this.loggifyConst(insn, "CONST  " + PrettyPrinter.toString(((LdcInsnNode) insn).cst));
                    break;

                case ILOAD:
                case LLOAD:
                case FLOAD:
                case DLOAD:
                case ALOAD:
                    this.loggifyLoad((VarInsnNode) insn);
                    break;

                case IALOAD:
                case LALOAD:
                case FALOAD:
                case DALOAD:
                case AALOAD:
                case BALOAD:
                case CALOAD:
                case SALOAD:
                    this.loggifyAload(insn);
                    break;

                case ISTORE:
                case LSTORE:
                case FSTORE:
                case DSTORE:
                case ASTORE:
                    this.loggifyStore((VarInsnNode) insn);
                    break;

                case IASTORE:
                case LASTORE:
                case FASTORE:
                case DASTORE:
                case AASTORE:
                case BASTORE:
                case CASTORE:
                case SASTORE:
                    this.loggifyAstore(insn);
                    break;

                case IADD:
                case LADD:
                case FADD:
                case DADD:
                    this.loggifyArith(insn, "ARITH  {0} + {1} = {2}");
                    break;
                case ISUB:
                case LSUB:
                case FSUB:
                case DSUB:
                    this.loggifyArith(insn, "ARITH  {0} - {1} = {2}");
                    break;
                case IMUL:
                case LMUL:
                case FMUL:
                case DMUL:
                    this.loggifyArith(insn, "ARITH  {0} * {1} = {2}");
                    break;
                case IDIV:
                case LDIV:
                case FDIV:
                case DDIV:
                    this.loggifyArith(insn, "ARITH  {0} / {1} = {2}");
                    break;
                case IREM:
                case LREM:
                case FREM:
                case DREM:
                    this.loggifyArith(insn, "ARITH  {0} % {1} = {2}");
                    break;
                case IAND:
                case LAND:
                    this.loggifyArith(insn, "ARITH  {0} & {1} = {2}");
                    break;
                case IOR:
                case LOR:
                    this.loggifyArith(insn, "ARITH  {0} | {1} = {2}");
                    break;
                case IXOR:
                case LXOR:
                    this.loggifyArith(insn, "ARITH  {0} ^ {1} = {2}");
                    break;
                case LCMP:
                case FCMPL:
                case DCMPL:
                    this.loggifyArith(insn, "ARITH  {0} <=> {1} = {2}");
                    break;
                case FCMPG:
                case DCMPG:
                    this.loggifyArith(insn, "ARITH  {0} <==> {1} = {2}");
                    break;
                case INEG:
                case LNEG:
                case FNEG:
                case DNEG:
                    this.loggifyArith(insn, "ARITH  -({0}) => {1}");
                    break;
                case ISHL:
                case LSHL:
                    this.loggifyArith(insn, "ARITH  {0} << {1} = {2}");
                    break;
                case ISHR:
                case LSHR:
                    this.loggifyArith(insn, "ARITH  {0} >> {1} = {2}");
                    break;
                case IUSHR:
                case LUSHR:
                    this.loggifyArith(insn, "ARITH  {0} >>> {1} = {2}");
                    break;
                case IF_ICMPEQ:
                case IF_ACMPEQ:
                    this.loggifyArith(insn, "ARITH  {0} == {1}");
                    break;
                case IF_ICMPNE:
                case IF_ACMPNE:
                    this.loggifyArith(insn, "ARITH  {0} != {1}");
                    break;
                case IF_ICMPLT:
                    this.loggifyArith(insn, "ARITH  {0} < {1}");
                    break;
                case IF_ICMPGE:
                    this.loggifyArith(insn, "ARITH  {0} >= {1}");
                    break;
                case IF_ICMPGT:
                    this.loggifyArith(insn, "ARITH  {0} > {1}");
                    break;
                case IF_ICMPLE:
                    this.loggifyArith(insn, "ARITH  {0} <= {1}");
                    break;

                case IINC:
                    this.loggifyIinc((IincInsnNode) insn);
                    break;

                case I2L:
                case I2F:
                case I2D:
                case I2B:
                case I2C:
                case I2S:
                case L2I:
                case L2F:
                case L2D:
                case F2I:
                case F2L:
                case F2D:
                case D2I:
                case D2L:
                case D2F:
                    this.loggifyConvert(insn);
                    break;

                case TABLESWITCH:
                case LOOKUPSWITCH:
                    this.loggifySwitch(insn);
                    break;

                case IRETURN:
                case LRETURN:
                case FRETURN:
                case DRETURN:
                case ARETURN:
                    this.loggifyReturn(insn, "RETURN {0}");
                    break;
                case RETURN:
                    this.loggifyReturn(insn, "RETURN");
                    break;

                case GETSTATIC: this.loggifyGetStatic((FieldInsnNode) insn); break;
                case GETFIELD:  this.loggifyGetField((FieldInsnNode) insn);  break;
                case PUTSTATIC: this.loggifyPutStatic((FieldInsnNode) insn); break;
                case PUTFIELD:  this.loggifyPutField((FieldInsnNode) insn);  break;

                case INVOKEVIRTUAL:
                case INVOKESPECIAL:
                case INVOKESTATIC:
                case INVOKEINTERFACE:
                    this.loggifyInvoke((MethodInsnNode) insn, opcode);
                    break;

                case NEW:            this.loggifyNew((TypeInsnNode) insn);                      break;
                case NEWARRAY:       this.loggifyNewarray((IntInsnNode) insn);                  break;
                case ANEWARRAY:      this.loggifyAnewarray((TypeInsnNode) insn);                break;
                case MULTIANEWARRAY: this.loggifyMultianewarray((MultiANewArrayInsnNode) insn); break;

                case ARRAYLENGTH: this.loggifyArraylength(insn); break;

                case ATHROW: this.loggifyAthrow(insn); break;

                case CHECKCAST: this.loggifyCheckcast((TypeInsnNode) insn); break;

                case INSTANCEOF: this.loggifyInstanceof((TypeInsnNode) insn); break;

                default:
                    // All other opcodes are not worth being logged.
                    ;
                    break;
                }
            }

            private void
            loggifyConst(final AbstractInsnNode insn, String message) {
                this.loggify(insn, Action.CONST, message);
            }

            private void
            loggifyLoad(final VarInsnNode vin) {
                InsnDescription id = OpcodeUtil.getInsnDescription(vin.getOpcode());

                String localVariable;
                {
                    LocalVariableNode lvn = this.localVariableFinder.getLocalVariable(vin.var);
                    localVariable = lvn == null ? "v" + vin.var : lvn.name;
                }

                this.loggify(
                    vin,
                    Action.LOAD,
                    localVariable,
                    vin.var,
                    id.resultOperandType,
                    "LOAD   @"  + vin.var + " => {0}"
                );
            }

            private void
            loggifyAload(final AbstractInsnNode insn) {
                InsnDescription id = OpcodeUtil.getInsnDescription(insn.getOpcode());
                this.loggify(
                    insn,
                    Action.ALOAD,
                    id.argumentOperandTypes,
                    "ALOAD  {0}[{1}] => {2}",
                    id.resultOperandType
                );
            }

            private void
            loggifyStore(VarInsnNode vin) {
                InsnDescription id = OpcodeUtil.getInsnDescription(vin.getOpcode());

                String localVariable;
                {
                    LocalVariableNode lvn = this.localVariableFinder.getLocalVariable(vin.var);
                    localVariable = lvn == null ? "v" + vin.var : lvn.name;
                }

                this.loggify(
                    vin,                        // target
                    Action.STORE,               // action
                    localVariable,              // custom
                    id.argumentOperandTypes,    // argumentOperandTypes
                    "STORE  {0} => @" + vin.var // message
                );
            }

            private void
            loggifyAstore(final AbstractInsnNode insn) {
                InsnDescription id = OpcodeUtil.getInsnDescription(insn.getOpcode());
                this.loggify(
                    insn,                    // target
                    Action.ASTORE,           // action
                    null,                    // custom
                    id.argumentOperandTypes, // argumentOperandTypes
                    "ASTORE {2} => {0}[{1}]" // message
                );
            }

            private void
            loggifyArith(AbstractInsnNode insn, String message) {
                InsnDescription id = OpcodeUtil.getInsnDescription(insn.getOpcode());
                if (id.resultOperandType == null) {
                    this.loggify(
                        insn,
                        Action.ARITH,
                        null,
                        id.argumentOperandTypes,
                        message
                    );
                } else {
                    this.loggify(
                        insn,
                        Action.ARITH,
                        id.argumentOperandTypes,
                        message,
                        id.resultOperandType
                    );
                }
            }

            private void
            loggifyIinc(IincInsnNode iin) {

                String localVariable;
                {
                    LocalVariableNode lvn = this.localVariableFinder.getLocalVariable(iin.var);
                    localVariable = lvn == null ? "v" + iin.var : lvn.name;
                }

                this.loggify(
                    iin,                             // target
                    Action.LOAD,                     // action
                    localVariable,                   // custom
                    iin.var,                         // localVariableIndex
                    Type.INT_TYPE,                   // localVariableType
                    "LOAD   @" + iin.var + " => {0}" // message
                );
                this.loggify(
                    iin,                         // target
                    Action.STORE,                // action
                    localVariable,               // custom
                    "STORE  {0} => @" + iin.var, // message
                    iin.var,                     // localVariableIndex
                    Type.INT_TYPE                // localVariableType
                );
                this.loggify(
                    iin,                                   // target
                    Action.ARITH,                          // action
                    iin.var,                               // localVariableIndex1
                    Type.INT_TYPE,                         // localVariableType1
                    "ARITH  {0} + " + iin.incr + " = {1}", // message
                    iin.var,                               // localVariableIndex2
                    Type.INT_TYPE                          // localVariableType2
                );
            }

            private void
            loggifyConvert(AbstractInsnNode insn) {
                InsnDescription id = OpcodeUtil.getInsnDescription(insn.getOpcode());
                this.loggify(
                    insn,
                    Action.CONVERT,
                    id.argumentOperandTypes,
                    "CONVRT {0} => {1}",
                    id.resultOperandType
                );
            }

            private void
            loggifySwitch(final AbstractInsnNode insn) {
                InsnDescription id = OpcodeUtil.getInsnDescription(insn.getOpcode());
                this.loggify(insn, Action.SWITCH, null, id.argumentOperandTypes, "SWITCH {0}");
            }

            private void
            loggifyReturn(AbstractInsnNode insn, String message) {
                InsnDescription id = OpcodeUtil.getInsnDescription(insn.getOpcode());
                if (id.argumentOperandTypes == null) {
                    this.loggify(insn, Action.RETURN, message);
                } else {
                    this.loggify(insn, Action.RETURN, null, id.argumentOperandTypes, message);
                }
            }

            private void
            loggifyGetStatic(final FieldInsnNode fin) {
                String className = Type.getObjectType(fin.owner).getClassName();
                this.instructions.insertBefore(fin, this.generateLogging(
                    (                                                         // loggerName
                        this.getLoggerName(Action.GET)
                        + '.'
                        + LoggifyingClassAdapter.dequalify(className)
                        + '.'
                        + fin.name
                    ),
                    this.getLogLevel(Action.GET, className + '.' + fin.name), // logLevel
                    (                                                         // message
                        "GET    "
                        + LoggifyingClassAdapter.dequalify(className)
                        + '.'
                        + fin.name
                        + " => {0}"
                    ),
                    InsnUtil.oa(                                              // pushParameters
                        this.wrapPrimitive(InsnUtil.il(fin.clone(LoggifyingClassAdapter.IDENTITY)), fin.desc)
                    )
                ));
            }

            private void
            loggifyGetField(final FieldInsnNode fin) {
                String className = Type.getObjectType(fin.owner).getClassName();
                this.instructions.insertBefore(fin, this.generateLogging(
                    (                                                             // loggerName
                        this.getLoggerName(Action.GET)
                        + '.'
                        + LoggifyingClassAdapter.dequalify(className)
                        + '.'
                        + fin.name
                    ),
                    this.getLogLevel(Action.GET, className + '.' + fin.name),     // logLevel
                    InsnUtil.il(new VarInsnNode(Opcodes.ASTORE, this.maxLocals)), // optionalPrefix
                    (                                                             // message
                        "GET    "
                        + LoggifyingClassAdapter.dequalify(className)
                        + '.'
                        + fin.name
                        + " (obj={0}) => {1}"
                    ),
                    InsnUtil.oa(                                                  // pushParameters
                        InsnUtil.il(new VarInsnNode(Opcodes.ALOAD, this.maxLocals)),
                        this.wrapPrimitive(InsnUtil.il(
                            new VarInsnNode(Opcodes.ALOAD, this.maxLocals),
                            fin.clone(LoggifyingClassAdapter.IDENTITY)
                        ), fin.desc)
                    ),
                    InsnUtil.il(new VarInsnNode(Opcodes.ALOAD, this.maxLocals))   // optionalSuffix
                ));
            }

            private void
            loggifyPutStatic(final FieldInsnNode fin) {
                String className = Type.getObjectType(fin.owner).getClassName();
                this.instructions.insert(fin, this.generateLogging(
                    (                                                                                // loggerName
                        this.getLoggerName(Action.PUT)
                        + '.'
                        + LoggifyingClassAdapter.dequalify(className)
                        + "."
                        + fin.name
                    ),
                    this.getLogLevel(Action.PUT, className + "." + fin.name),                        // logLevel
                    "PUT    {0} => " + LoggifyingClassAdapter.dequalify(className) + "." + fin.name, // message
                    InsnUtil.oa(this.wrapPrimitive(InsnUtil.il(                                      // pushParameters
                        new FieldInsnNode(Opcodes.GETSTATIC, fin.owner, fin.name, fin.desc)
                    ), fin.desc))
                ));
            }

            private void
            loggifyPutField(final FieldInsnNode fin) {
                String className = Type.getObjectType(fin.owner).getClassName();
                this.instructions.insertBefore(fin, this.generateLoggingWithOperands(
                    (                                                         // loggerName
                        this.getLoggerName(Action.PUT)
                        + '.'
                        + LoggifyingClassAdapter.dequalify(className)
                        + "."
                        + fin.name
                    ),
                    this.getLogLevel(Action.PUT, className + "." + fin.name), // logLevel
                    (                                                         // message
                        "PUT    {1} => "
                        + LoggifyingClassAdapter.dequalify(className)
                        + "."
                        + fin.name
                        + " (obj={0})"
                    ),
                    new Type[] {                                              // operandTypes
                        Type.getObjectType(fin.owner),
                        Type.getType(fin.desc),
                    }
                ));
            }

            private void
            loggifyInvoke(final MethodInsnNode min, int opcode) {
                String calledMethodName       = min.name;
                String calledMethodDescriptor = min.desc;
                String calledMethodOwner      = min.owner;

                LOGGIFY_INVOKE: {
                    Level logLevel = this.getLogLevel(Action.INVOKE, null);
                    if (logLevel == Level.OFF) break LOGGIFY_INVOKE;
                    Type[]        operandTypes;
                    int           idx    = 0;
                    StringBuilder format = (
                        new StringBuilder("INVOKE ")
                        .append(LoggifyingClassAdapter.dequalify(Signature.toString(calledMethodOwner, null)))
                        .append('.')
                        .append(LoggifyingClassAdapter.dequalify(
                            Signature.toString(calledMethodName, calledMethodDescriptor, null)
                        ))
                        .append(": ")
                    );
                    Type[] parameterTypes = Type.getArgumentTypes(calledMethodDescriptor);
                    if (opcode != Opcodes.INVOKESTATIC && !"<init>".equals(calledMethodName)) {
                        operandTypes        = new Type[1 + parameterTypes.length];
                        operandTypes[idx++] = Type.getObjectType(calledMethodOwner);
                        format.append("target={0}, ");
                    } else {
                        operandTypes = new Type[parameterTypes.length];
                    }
                    format.append("args=(");
                    for (int i = 0; i < parameterTypes.length; i++) {
                        Type parameterType = parameterTypes[i];
                        format.append(i == 0 ? "{" + idx + "}" : ", {" + idx + "}");
                        operandTypes[idx++] = parameterType;
                    }
                    format.append(')');
                    this.instructions.insertBefore(min, this.generateLoggingWithOperands(
                        this.getLoggerName(Action.INVOKE), // loggerName
                        logLevel,                          // logLevel
                        format.toString(),                 // message
                        operandTypes                       // operandTypes
                    ));
                }

                Type returnType = Type.getReturnType(calledMethodDescriptor);
                this.instructions.insert(min, this.generateLoggingWithOperands(
                    this.getLoggerName(Action.RESULT),                                     // loggerName
                    this.getLogLevel(Action.RESULT, null),                                 // level
                    (                                                                      // message
                        "RESULT "
                        + LoggifyingClassAdapter.dequalify(Signature.toString(calledMethodOwner, null))
                        + '.'
                        + LoggifyingClassAdapter.dequalify(
                            Signature.toString(calledMethodName, calledMethodDescriptor, null)
                        )
                        + (returnType == Type.VOID_TYPE ? "" : " => {0}")
                    ),
                    returnType == Type.VOID_TYPE ? new Type[0] : new Type[] { returnType } // operandTypes
                ));
            }

            private void
            loggifyNew(TypeInsnNode tin) {
                String className = Type.getObjectType(tin.desc).getClassName();
                this.instructions.insert(tin, this.generateLogging(
                    this.getLoggerName(Action.NEW) + '.' + LoggifyingClassAdapter.dequalify(className), // loggerName
                    this.getLogLevel(Action.NEW, className),                                            // logLevel
                    "NEW    " + LoggifyingClassAdapter.dequalify(className)                             // message
                ));
            }

            private void
            loggifyNewarray(IntInsnNode iin) {
                @SuppressWarnings("synthetic-access") String atype = LoggifyingClassAdapter.decodeAtype(iin.operand);
                this.instructions.insertBefore(iin, this.generateLoggingWithOperands(
                    (                                            // loggerName
                        this.getLoggerName(Action.NEW)
                        + '.'
                        + LoggifyingClassAdapter.dequalify(atype)
                        + "[x]"
                    ),
                    this.getLogLevel(Action.NEW, atype + "[x]"), // logLevel
                    "NEW    " + atype + "[{0}]",                 // message
                    new Type[] { Type.INT_TYPE }                 // operandTypes
                ));
            }

            private void
            loggifyAnewarray(TypeInsnNode tin) {
                String className = Type.getObjectType(tin.desc).getClassName();
                this.instructions.insertBefore(tin, this.generateLoggingWithOperands(
                    (                                                                  // loggerName
                        this.getLoggerName(Action.NEW)
                        + '.'
                        + LoggifyingClassAdapter.dequalify(className)
                        + "[x]"
                    ),
                    this.getLogLevel(Action.NEW, className + "[x]"),                   // logLevel
                    "NEW    " + LoggifyingClassAdapter.dequalify(className) + "[{0}]", // message
                    new Type[] { Type.INT_TYPE }                                       // operandTypes
                ));
            }

            private void
            loggifyArraylength(final AbstractInsnNode insn) {
                this.instructions.insertBefore(insn, this.generateLoggingWithOperands(
                    this.getLoggerName(Action.LENGTH),     // loggerName
                    this.getLogLevel(Action.LENGTH, null), // logLevel
                    "LENGTH {0}",                          // message
                    new Type[] { Types.OBJECT_TYPE }       // operandTypes
                ));
            }

            private void
            loggifyMultianewarray(MultiANewArrayInsnNode manain) {
                String     suffix;
                String     message;
                List<Type> operandTypes;
                String     elementClassName;
                {
                    Type arrayType = Type.getObjectType(manain.desc);
                    elementClassName = arrayType.getElementType().getClassName();

                    suffix       = "";
                    message      = "NEW    " + LoggifyingClassAdapter.dequalify(elementClassName);
                    operandTypes = new ArrayList<Type>();
                    int i;
                    for (i = 0; i < manain.dims; i++) {
                        suffix  += "[x]";
                        message += "[{" + i + "}]";
                        operandTypes.add(Type.INT_TYPE);
                    }
                    int totalDimensions = arrayType.getDimensions();
                    for (; i < totalDimensions; i++) {
                        suffix  += "[]";
                        message += "[]";
                    }
                }

                this.instructions.insertBefore(manain, this.generateLoggingWithOperands(
                    (                                                        // loggerName
                        this.getLoggerName(Action.NEW)
                        + '.'
                        + LoggifyingClassAdapter.dequalify(elementClassName)
                        + suffix
                    ),
                    this.getLogLevel(Action.NEW, elementClassName + suffix), // logLevel
                    message,                                                 // message
                    operandTypes.toArray(new Type[operandTypes.size()])      // operandTypes
                ));
            }

            private void
            loggifyAthrow(final AbstractInsnNode insn) {
                int tmpVar = this.maxLocals;
                this.instructions.insertBefore(
                    insn,
                    LoggifyingClassAdapter.this.generateLoggingWithThrowable(
                        this.getLoggerName(Action.THROW),                     // loggerName
                        this.getLogLevel(Action.THROW, null),                 // logLevel
                        InsnUtil.il(new VarInsnNode(Opcodes.ASTORE, tmpVar)), // optionalPrefix
                        this.sourceMethodName,                                // sourceMethodName
                        this.currentLineProvider.getCurrentLine(),            // line
                        "THROW",                                              // message
                        InsnUtil.il(new VarInsnNode(Opcodes.ALOAD, tmpVar)),  // pushThrowable
                        InsnUtil.il(new VarInsnNode(Opcodes.ALOAD, tmpVar)),  // optionalSuffix
                        this                                                  // methodVisitor
                    )
                );
            }

            private void
            loggifyCheckcast(TypeInsnNode tin) {
                String className = Type.getObjectType(tin.desc).getClassName();
                this.instructions.insertBefore(tin, this.generateLoggingWithOperands(
                    this.getLoggerName(Action.CAST) + '.' + LoggifyingClassAdapter.dequalify(className), // loggerName
                    this.getLogLevel(Action.CAST, className),                                            // logLevel
                    "CAST   " + LoggifyingClassAdapter.dequalify(className),                             // message
                    new Type[] { Types.OBJECT_TYPE }                                                     // operandTypes
                ));
            }

            private void
            loggifyInstanceof(TypeInsnNode tin) {
                String className = Type.getObjectType(tin.desc).getClassName();
                this.instructions.insertBefore(tin, this.generateLoggingWithOperands(
                    (                                                        // loggerName
                        this.getLoggerName(Action.INSTANCEOF)
                        + '.'
                        + LoggifyingClassAdapter.dequalify(className)
                    ),
                    this.getLogLevel(Action.INSTANCEOF, className),          // logLevel
                    "INSTOF " + LoggifyingClassAdapter.dequalify(className), // message
                    new Type[] { Types.OBJECT_TYPE }                         // operandTypes
                ));
            }

            /**
             * Replaces the {@code target} instruction with code that checks "isLoggable()", logs the instruction, and
             * then executes the instruction.
             */
            private void
            loggify(AbstractInsnNode target, Action action, String message) {
                Level logLevel = this.getLogLevel(action, null);
                if (logLevel == Level.OFF) return;
                String loggerName = this.getLoggerName(action);

                InsnList insns = this.generateLogpLocalVariables(
                    loggerName,  // loggerName
                    logLevel,    // logLevel
                    message,     // message
                    0,           // firstLocalVariableIndex
                    new Type[0], // localVariableTypes
                    false        // withNames
                );

                this.instructions.insertBefore(
                    target,
                    LoggifyingClassAdapter.this.generateIsLoggable(loggerName, logLevel, insns, null)
                );
            }

            /**
             * Replaces the {@code target} instruction with code that checks "isLoggable()", executes the instruction,
             * and then logs the instruction with its argument operands and result operand.
             */
            private void
            loggify(
                AbstractInsnNode target,
                Action           action,
                Type[]           argumentOperandTypes,
                String           message,
                Type             resultOperandType
            ) {
                Level logLevel = this.getLogLevel(action, null);
                if (logLevel == Level.OFF) return;
                String loggerName = this.getLoggerName(action);

                Type[] resultOperandTypes = new Type[] { resultOperandType };

                InsnList insns1;
                {
                    int[] idx = new int[] { this.maxLocals };
                    insns1 = this.generateStoreOperands(argumentOperandTypes, idx);
                    insns1.add(this.generateLoadOperands(argumentOperandTypes, idx[0]));
                    insns1.add(target.clone(LoggifyingClassAdapter.IDENTITY));
                    insns1.add(this.generateStoreOperands(resultOperandTypes, idx));
                    insns1.add(this.generateLogpLocalVariables(
                        loggerName,                                          // loggerName
                        logLevel,                                            // logLevel
                        message,                                             // message
                        this.maxLocals,                                      // firstLocalVariableIndex
                        this.join(argumentOperandTypes, resultOperandTypes), // localVariableTypes
                        false                                                // withNames
                    ));
                    insns1.add(this.generateLoadOperands(resultOperandTypes, idx[0]));
                }

                InsnList insns2 = InsnUtil.il(target.clone(LoggifyingClassAdapter.IDENTITY));

                this.instructions.insert(
                    target,
                    LoggifyingClassAdapter.this.generateIsLoggable(loggerName, logLevel, insns1, insns2)
                );
                this.instructions.remove(target);
            }

            /**
             * Replaces the {@code target} instruction with code that checks "isLoggable()", logs the instruction with
             * its argument operands, and then executes the instruction.
             */
            private void
            loggify(
                AbstractInsnNode target,
                Action           action,
                String           custom,
                Type[]           argumentOperandTypes,
                String           message
            ) {
                Level logLevel = this.getLogLevel(action, custom);
                if (logLevel == Level.OFF) return;
                String loggerName = this.getLoggerName(action);
                if (custom != null) loggerName += '.' + custom;

                InsnList insns;
                {
                    int[] idx = new int[] { this.maxLocals };
                    insns = this.generateStoreOperands(argumentOperandTypes, idx);
                    insns.add(this.generateLogpLocalVariables(
                        loggerName,           // loggerName
                        logLevel,             // logLevel
                        message,              // message
                        this.maxLocals,       // firstLocalVariableIndex
                        argumentOperandTypes, // localVariableTypes
                        false                 // withNames
                    ));
                    insns.add(this.generateLoadOperands(argumentOperandTypes, idx[0]));
                }

                this.instructions.insertBefore(
                    target,
                    LoggifyingClassAdapter.this.generateIsLoggable(loggerName, logLevel, insns, null)
                );
            }

            /**
             * Checks "isLoggable()", logs the given local variable, and then executes the instruction.
             * @param custom
             */
            private void
            loggify(
                AbstractInsnNode target,
                Action           action,
                String           custom,
                int              localVariableIndex,
                Type             localVariableType,
                String           message
            ) {
                Level logLevel = this.getLogLevel(action, custom);
                if (logLevel == Level.OFF) return;
                String loggerName = this.getLoggerName(action) + '.' + custom;

                InsnList insns = this.generateLogpLocalVariables(
                    loggerName,                       // loggerName
                    logLevel,                         // logLevel
                    message,                          // message
                    localVariableIndex,               // firstLocalVariableIndex
                    new Type[] { localVariableType }, // localVariableTypes
                    false                             // withNames
                );

                this.instructions.insertBefore(
                    target,
                    LoggifyingClassAdapter.this.generateIsLoggable(loggerName, logLevel, insns, null)
                );
            }

            /**
             * Executes the instruction, then checks "isLoggable()", and then logs the given local variable.
             */
            private void
            loggify(
                AbstractInsnNode target,
                Action           action,
                String           custom,
                String           message,
                int              localVariableIndex,
                Type             localVariableType
            ) {
                Level logLevel = this.getLogLevel(action, custom);
                if (logLevel == Level.OFF) return;
                String loggerName = this.getLoggerName(action);
                if (custom != null) loggerName += '.' + custom;

                InsnList insns = this.generateLogpLocalVariables(
                    loggerName,                       // loggerName
                    logLevel,                         // logLevel
                    message,                          // message
                    localVariableIndex,               // firstLocalVariableIndex
                    new Type[] { localVariableType }, // localVariableTypes
                    false                             // withNames
                );

                this.instructions.insert(
                    target,
                    LoggifyingClassAdapter.this.generateIsLoggable(loggerName, logLevel, insns, null)
                );
            }

            /**
             * Checks "isLoggable()", then stores {@code var1}, then executes the instruction, and then logs the
             * stored value of {@code var1} and {@code var2}.
             */
            private void
            loggify(
                AbstractInsnNode target,
                Action           action,
                int              localVariableIndex1,
                Type             localVariableType1,
                String           message,
                int              localVariableIndex2,
                Type             localVariableType2
            ) {
                Level logLevel = this.getLogLevel(action, null);
                if (logLevel == Level.OFF) return;
                String loggerName = this.getLoggerName(action);

                InsnList insns1 = new InsnList();
                insns1.add(new VarInsnNode(localVariableType1.getOpcode(Opcodes.ILOAD), localVariableIndex1));
                insns1.add(new VarInsnNode(localVariableType1.getOpcode(Opcodes.ISTORE), this.maxLocals));
                insns1.add(target.clone(LoggifyingClassAdapter.IDENTITY));
                insns1.add(this.generateLogpLocalVariables(
                    loggerName,                                            // loggerName
                    logLevel,                                              // logLevel
                    message,                                               // message
                    new int[] { this.maxLocals, localVariableIndex2 },     // localVariableIndexes
                    new Type[] { localVariableType1, localVariableType2 }, // localVariableTypes
                    false                                                  // withNames
                ));

                InsnList insns2 = InsnUtil.il(target.clone(LoggifyingClassAdapter.IDENTITY));

                this.instructions.insert(
                    target,
                    LoggifyingClassAdapter.this.generateIsLoggable(loggerName, logLevel, insns1, insns2)
                );
                this.instructions.remove(target);
            }

            private InsnList
            generateLogpLocalVariables(
                String  loggerName,
                Level   logLevel,
                String  message,
                int     firstLocalVariableIndex,
                Type[]  localVariableTypes,
                boolean withNames
            ) {
                int[] localVariableIndexes = new int[localVariableTypes.length];
                for (int i = 0; i < localVariableIndexes.length; i++) {
                    localVariableIndexes[i] = firstLocalVariableIndex;
                    firstLocalVariableIndex += localVariableTypes[i].getSize();
                }
                return this.generateLogpLocalVariables(
                    loggerName,
                    logLevel,
                    message,
                    localVariableIndexes,
                    localVariableTypes,
                    withNames
                );
            }

            private InsnList
            generateLogpLocalVariables(
                String  loggerName,
                Level   logLevel,
                String  message,
                int[]   localVariableIndexes,
                Type[]  localVariableTypes,
                boolean withNames
            ) {
                List<InsnList> pushParameters = new ArrayList<InsnList>();
                for (int i = 0; i < localVariableTypes.length; i++) {
                    Type localVariableType = localVariableTypes[i];
                    assert localVariableType != Type.VOID_TYPE;
                    int localVariableIndex = localVariableIndexes[i];

                    if (withNames) {
                        LocalVariableNode lv = this.localVariableFinder.getLocalVariable(localVariableIndex);
                        if (lv != null && lv.name != null) {
                            message = message.replace("{" + i + "}", lv.name + "={" + i + "}");
                        }
                    }

                    pushParameters.add(this.wrapPrimitive(InsnUtil.il(
                        new VarInsnNode(localVariableType.getOpcode(Opcodes.ILOAD), localVariableIndex)
                    ), localVariableType.getDescriptor()));
                }

                // Replace occurrences of "@<lvidx>" with the type and name of the local variable.
                for (;;) {
                    int at = message.indexOf('@');
                    if (at == -1) break;
                    int beg = at + 1, end;
                    for (end = beg; end < message.length() && Character.isDigit(message.charAt(end)); end++);
                    if (beg == end) break;
                    int localVariableIndex = Integer.parseInt(message.substring(beg, end));

                    LocalVariableNode lvn        = this.localVariableFinder.getLocalVariable(localVariableIndex);
                    String            substitute = (
                        lvn == null ? "v" + localVariableIndex :
                        "this".equals(lvn.name) ? "this" :
                        LoggifyingClassAdapter.dequalify(Type.getType(lvn.desc).getClassName()) + " " + lvn.name
                    );
                    message = message.substring(0, at) + substitute + message.substring(end);
                }

                InsnList il = new InsnList();

                LoggifyingClassAdapter.this.loggingCount++;

                // Helper.logp(
                //     this-class-name.logger-field-name, // logger
                //     Level.log-level-name,              // level
                //     source-class-name,                 // sourceClassName
                //     source-method-name,                // sourceMethodName
                //     source,                            // source
                //     lineNumber,                        // lineNumber
                //     message,                           // message
                //     push-parameters                    // parameters
                // );

                // Logger logger
                String loggerFieldName = LoggifyingClassAdapter.loggerNameToFieldName(loggerName);
                il.add(new FieldInsnNode(
                    Opcodes.GETSTATIC,
                    LoggifyingClassAdapter.this.thisClassName,
                    loggerFieldName,
                    "Ljava/util/logging/Logger;"
                ));
                // Level level
                il.add(LoggifyingClassAdapter.generatePushLogLevel(logLevel));
                // String sourceClassName
                il.add(
                    LoggifyingClassAdapter.this.sourceClassName == null
                    ? new InsnNode(Opcodes.ACONST_NULL)
                    : new LdcInsnNode(LoggifyingClassAdapter.this.sourceClassName)
                );
                // String sourceMethodName
                il.add(
                    this.sourceMethodName == null
                    ? new InsnNode(Opcodes.ACONST_NULL)
                    : new LdcInsnNode(this.sourceMethodName)
                );
                // String source
                il.add(
                    LoggifyingClassAdapter.this.source == null
                    ? new InsnNode(Opcodes.ACONST_NULL)
                    : new LdcInsnNode(LoggifyingClassAdapter.this.source)
                );
                // int  lineNumber
                il.add(new LdcInsnNode(this.currentLineProvider.getCurrentLine()));
                // String message
                il.add(new LdcInsnNode(message));
                // Object[] parameters
                il.add(InsnUtil.oa(pushParameters.toArray(new InsnList[pushParameters.size()])));

                il.add(new MethodInsnNode(Opcodes.INVOKESTATIC, LoggifyingClassAdapter.HELPER_CLASS_NAME, "logp", (
                    "("
                    + "Ljava/util/logging/Logger;"  // Logger   logger
                    + "Ljava/util/logging/Level;"   // Level    level
                    + "Ljava/lang/String;"          // String   sourceClassName
                    + "Ljava/lang/String;"          // String   sourceMethodName
                    + "Ljava/lang/String;"          // String   source
                    + "I"                           // int      lineNumber
                    + "Ljava/lang/String;"          // String   message
                    + "[Ljava/lang/Object;"         // Object[] parameters
                    + ")V"
                )));

                return il;
            }

            private Type[]
            join(Type[] a1, Type[] a2) {
                Type[] result = new Type[a1.length + a2.length];
                System.arraycopy(a1, 0, result, 0, a1.length);
                System.arraycopy(a2, 0, result, a1.length, a2.length);
                return result;
            }

            private InsnList
            generateStoreOperands(Type[] operandTypes, int[] idx) {
                InsnList il = new InsnList();
                for (Type operandType : operandTypes) {
                    assert operandType != Type.VOID_TYPE;
                    VarInsnNode insn = new VarInsnNode(operandType.getOpcode(Opcodes.ISTORE), idx[0]);
                    if (il.size() == 0) {
                        il.add(insn);
                    } else {
                        il.insertBefore(il.getFirst(), insn);
                    }
                    idx[0] += operandType.getSize();
                }
                return il;
            }

            private InsnList
            generateLoadOperands(Type[] operandTypes, int idx) {
                InsnList il = new InsnList();
                for (int i = operandTypes.length - 1; i >= 0; i--) {
                    Type operandType = operandTypes[i];
                    idx -= operandType.getSize();
                    VarInsnNode insn = new VarInsnNode(operandType.getOpcode(Opcodes.ILOAD), idx);
                    if (il.size() == 0) {
                        il.add(insn);
                    } else {
                        il.insertBefore(il.getFirst(), insn);
                    }
                }
                return il;
            }

            /**
             * If {@code desc} denotes a primitive type, then {@code il} is augmented with instructions that convert
             * the primitive operand on top of the operand stack into a wrapper class object reference.
             */
            private InsnList
            wrapPrimitive(InsnList il, String desc) {
                if ("Z".equals(desc)) {
                    il.add(new MethodInsnNode(
                        Opcodes.INVOKESTATIC,
                        "java/lang/Boolean",
                        "valueOf",
                        "(Z)Ljava/lang/Boolean;"
                    ));
                } else
                if ("B".equals(desc)) {
                    il.add(new MethodInsnNode(
                        Opcodes.INVOKESTATIC,
                        "java/lang/Byte",
                        "valueOf",
                        "(B)Ljava/lang/Byte;"
                    ));
                } else
                if ("C".equals(desc)) {
                    il.add(new MethodInsnNode(
                        Opcodes.INVOKESTATIC,
                        "java/lang/Character",
                        "valueOf",
                        "(C)Ljava/lang/Character;"
                    ));
                } else
                if ("D".equals(desc)) {
                    il.add(new MethodInsnNode(
                        Opcodes.INVOKESTATIC,
                        "java/lang/Double",
                        "valueOf",
                        "(D)Ljava/lang/Double;"
                    ));
                } else
                if ("F".equals(desc)) {
                    il.add(new MethodInsnNode(
                        Opcodes.INVOKESTATIC,
                        "java/lang/Float",
                        "valueOf",
                        "(F)Ljava/lang/Float;"
                    ));
                } else
                if ("I".equals(desc)) {
                    il.add(new MethodInsnNode(
                        Opcodes.INVOKESTATIC,
                        "java/lang/Integer",
                        "valueOf",
                        "(I)Ljava/lang/Integer;"
                    ));
                } else
                if ("J".equals(desc)) {
                    il.add(new MethodInsnNode(
                        Opcodes.INVOKESTATIC,
                        "java/lang/Long",
                        "valueOf",
                        "(J)Ljava/lang/Long;"
                    ));
                } else
                if ("S".equals(desc)) {
                    il.add(new MethodInsnNode(
                        Opcodes.INVOKESTATIC,
                        "java/lang/Short",
                        "valueOf",
                        "(S)Ljava/lang/Short;"
                    ));
                }
                return il;
            }

            private Level
            getLogLevel(Action action, String custom) {
                String scn = LoggifyingClassAdapter.this.sourceClassName;
                assert scn != null;
                return LoggifyingMethodVisitor.this.methodLogLevelCalculator.calculate(
                    action,
                    LoggifyingClassAdapter.this.source,
                    LoggifyingClassAdapter.this.thisClassAccess,
                    scn,
                    methodAccess,
                    this.sourceMethodName,
                    this.currentLineProvider.getCurrentLine(),
                    custom
                );
            }

            /**
             * Generates logging code that reports the top values on the operand stack.
             * @param message '{<i>operandTypes.length - 1</i>}' refers to the top element of the operand stack,
             *                '{<i>operandTypes.length - 2</i>}' to the second-from-top operand, and so forth
             */
            private InsnList
            generateLoggingWithOperands(
                String loggerName,
                Level  logLevel,
                String message,
                Type[] operandTypes
            ) {
                if (logLevel == Level.OFF) return LoggifyingClassAdapter.NO_INSNS;

                InsnList prefix = new InsnList();
                InsnList suffix = new InsnList();
                int      idx    = this.maxLocals;
                for (Type operandType : operandTypes) {
                    assert operandType != Type.VOID_TYPE;
                    VarInsnNode insn = new VarInsnNode(operandType.getOpcode(Opcodes.ISTORE), idx);
                    if (prefix.size() == 0) {
                        prefix.add(insn);
                    } else {
                        prefix.insertBefore(prefix.getFirst(), insn);
                    }

                    suffix.add(new VarInsnNode(operandType.getOpcode(Opcodes.ILOAD), idx));
                    idx += operandType.getSize();
                }
                return this.generateLoggingWithLocalVariables(
                    loggerName,          // loggerName
                    logLevel,            // logLevel
                    prefix,              // optionalPrefix
                    message,             // message
                    this.maxLocals,      // firstLocalVariableIndex
                    operandTypes,        // localVariableTypes
                    suffix,              // optionalSuffix
                    false                // withNames
                );
            }

            /**
             * @see #generateLoggingWithLocalVariables(String, Level, InsnList, String, int, Type[], InsnList,
             *      boolean)
             */
            private InsnList
            generateLoggingWithLocalVariables(
                String  loggerName,
                Level   logLevel,
                String  message,
                int     firstLocalVariableIndex,
                Type[]  localVariableTypes,
                boolean withNames
            ) {
                return this.generateLoggingWithLocalVariables(
                    loggerName,
                    logLevel,
                    null,                    // optionalPrefix
                    message,                 // message
                    firstLocalVariableIndex,
                    localVariableTypes,
                    null,                    // optionalSuffix
                    withNames
                );
            }

            /**
             * Generates logging code that reports the values of local variables
             * @param message '{0}' refers to the local variable at {@code firstLocalVariableIndex}, '{1}' to the local
             *               variable at {@code firstLocalVariableIndex + 1} and so forth
             */
            private InsnList
            generateLoggingWithLocalVariables(
                String   loggerName,
                Level    logLevel,
                InsnList optionalPrefix,
                String   message,
                int      firstLocalVariableIndex,
                Type[]   localVariableTypes,
                InsnList optionalSuffix,
                boolean  withNames
            ) {
                if (logLevel == Level.OFF) return LoggifyingClassAdapter.NO_INSNS;

                int            idx            = firstLocalVariableIndex;
                List<InsnList> pushParameters = new ArrayList<InsnList>();
                for (int i = 0; i < localVariableTypes.length; i++) {
                    Type localVariableType = localVariableTypes[i];
                    assert localVariableType != Type.VOID_TYPE;

                    if (withNames) {
                        LocalVariableNode lv = this.localVariableFinder.getLocalVariable(idx);
                        if (lv != null && lv.name != null) {
                            message = message.replace("{" + i + "}", lv.name + "={" + i + "}");
                        }
                    }

                    pushParameters.add(this.wrapPrimitive(InsnUtil.il(
                        new VarInsnNode(localVariableType.getOpcode(Opcodes.ILOAD), idx)
                    ), localVariableType.getDescriptor()));

                    idx += localVariableType.getSize();
                }
                return this.generateLogging(
                    loggerName,
                    logLevel,
                    optionalPrefix,
                    message,
                    InsnUtil.oa(pushParameters.toArray(new InsnList[pushParameters.size()])),
                    optionalSuffix
                );
            }

            private InsnList
            generateLogging(String loggerName, Level logLevel, String message) {
                return this.generateLogging(
                    loggerName,
                    logLevel,
                    message,
                    InsnUtil.il(new InsnNode(Opcodes.ACONST_NULL))
                );
            }

            private InsnList
            generateLogging(String loggerName, Level logLevel, String message, InsnList pushParameters) {
                return this.generateLogging(loggerName, logLevel, null, message, pushParameters, null);
            }

            private InsnList
            generateLogging(
                String   loggerName,
                Level    logLevel,
                InsnList optionalPrefix,
                String   message,
                InsnList pushParameters,
                InsnList optionalSuffix
            ) {
                return LoggifyingClassAdapter.this.generateLogging(
                    loggerName,                                // loggerName
                    logLevel,                                  // logLevel
                    optionalPrefix,                            // optionalPrefix
                    this.sourceMethodName,                     // sourceMethodName
                    this.currentLineProvider.getCurrentLine(), // line
                    message,                                   // message
                    pushParameters,                            // pushParameters
                    optionalSuffix,                            // optionalSuffix
                    this                                       // methodVisitor
                );
            }

            private String
            getLoggerName(Action action) {

                String packagePrefix, unqualifiedTopLevelTypeName, nestedTypeSuffix;
                {
                    String scn = LoggifyingClassAdapter.this.thisClassName;
                    assert scn != null;
                    String[] sa = LoggifyingClassAdapter.splitClassName(scn);
                    packagePrefix               = sa[0];
                    unqualifiedTopLevelTypeName = sa[1];
                    nestedTypeSuffix            = sa[2];
                }

                // Determine return type name and parameter type names from signature or descriptor.
                String returnTypeName, parameterTypeNames;
                if (optionalMethodSignature != null) {
                    TraceSignatureVisitor tsv = new TraceSignatureVisitor(0);
                    new SignatureReader(optionalMethodSignature).accept(tsv);
                    returnTypeName     = tsv.getReturnType().toString();
                    parameterTypeNames = tsv.getDeclaration();
                } else {
                    returnTypeName = Type.getReturnType(methodDescriptor).getClassName();
                    Type[] ats = Type.getArgumentTypes(methodDescriptor);
                    if (ats.length == 0) {
                        parameterTypeNames = "()";
                    } else {
                        StringBuilder sb = new StringBuilder("(").append(ats[0].getClassName());
                        for (int i = 1; i < ats.length; ++i) {
                            sb.append(", ").append(ats[i].getClassName());
                        }
                        parameterTypeNames = sb.append(')').toString();
                    }
                }

                return LoggifyingClassAdapter.cookLoggerName(
                    LoggifyingClassAdapter.METHOD_LOGGER_NAME_FORMAT.format(new Object[] {
                        packagePrefix,
                        unqualifiedTopLevelTypeName,
                        nestedTypeSuffix,
                        LoggifyingClassAdapter.dequalify(returnTypeName),
                        methodName,
                        LoggifyingClassAdapter.dequalify(parameterTypeNames),
                        action,
                    })
                );
            }
        }
        return new LoggifyingMethodVisitor();
    }

    private static String
    decodeAtype(int operand) {
        switch (operand) {
        case T_BOOLEAN: return "boolean";
        case T_CHAR:    return "char";
        case T_FLOAT:   return "float";
        case T_DOUBLE:  return "double";
        case T_BYTE:    return "byte";
        case T_SHORT:   return "short";
        case T_INT:     return "int";
        case T_LONG:    return "long";
        default:        return "???" + operand + "???";
        }
    }

    /**
     * Later lines haver higher precendence.
     */
    private static LevelCalculator
    parseLoggifyAnnotation(String[] lines, LevelCalculator levelCalculator, final ErrorHandler errorHandler) {

        for (final String line : lines) {
            Matcher m = LoggifyingClassAdapter.LOGGIFY_ANNOTATION_PATTERN.matcher(line);
            if (!m.matches()) {
                errorHandler.handle(
                    "Rule \""
                    + line
                    + "\" has invalid value format; must be "
                    + "\"<action> [ ',' <action> ]... '=' <level> [ ':' <discriminator> ]\""
                );
                continue;
            }

            final EnumSet<Action> actions;
            {
                String[] actionNames = m.group(1).split("\\s*,\\s*");
                actions = EnumSet.noneOf(Action.class);
                for (String actionName : actionNames) {
                    try {
                        if ("ALL".equals(actionName)) {
                            actions.addAll(EnumSet.allOf(Action.class));
                        } else {
                            actions.add(Action.valueOf(actionName));
                        }
                    } catch (IllegalArgumentException iae) {
                        errorHandler.handle(
                            "Rule \""
                            + line
                            + "\": Invalid action '"
                            + actionName
                            + "'; allowed values are '"
                            + Arrays.toString(Action.values()) + "'"
                        );
                    }
                }
            }
            final Level level;
            {
                @NotNull String levelName = m.group(2);
                try {
                    level = Level.parse(levelName);
                } catch (IllegalArgumentException iae) {
                    errorHandler.handle("Rule \"" + line + "\": Invalid level '" + levelName + "'");
                    continue;
                }
            }

            @Nullable String discriminatorExpression = m.group(3);
            try {
                levelCalculator = LevelCalculatorUtil.discriminated(
                    actions,
                    level,
                    discriminatorExpression,
                    levelCalculator
                );
            } catch (ParseException pe) {
                errorHandler.handle(pe);
            }
        }

        return levelCalculator;
    }
    private static final Pattern LOGGIFY_ANNOTATION_PATTERN = Pattern.compile("([\\w, ]+)=(\\w+)(?::(.*))?");

    /**
     * Removes all qualifiers from all qualified names.
     * <p>
     * Example:
     * <p>
     * 'pkg.meth(java.lang.String)' => 'meth(String)'
     */
    private static String
    dequalify(String s) {
        return LoggifyingClassAdapter.DEQUALIFY.matcher(s).replaceAll("");
    }
    private static final Pattern DEQUALIFY = Pattern.compile("\\w+\\.");

    private String
    getLoggerName(Action action) {

        String packagePrefix, unqualifiedTopLevelTypeName, nestedTypeSuffix;
        {
            String tcn = this.thisClassName;
            assert tcn != null;
            String[] sa = LoggifyingClassAdapter.splitClassName(tcn);
            packagePrefix               = sa[0];
            unqualifiedTopLevelTypeName = sa[1];
            nestedTypeSuffix            = sa[2];
        }

        return LoggifyingClassAdapter.cookLoggerName(
            LoggifyingClassAdapter.CLASS_LOGGER_NAME_FORMAT.format(new Object[] {
                packagePrefix,               // {0}
                unqualifiedTopLevelTypeName, // {1}
                nestedTypeSuffix,            // {2}
                null,
                null,
                null,
                action                       // {6}
            })
        );
    }

    private static String[]
    splitClassName(String className) {
        int    idx1          = 1 + className.lastIndexOf('/');
        String packagePrefix = className.substring(0, idx1).replace('/', '.');
        int    idx2          = className.indexOf('$', idx1);
        String nestedTypeSuffix;
        String unqualifiedTopLevelTypeName;
        if (idx2 == -1) {
            unqualifiedTopLevelTypeName = className.substring(idx1);
            nestedTypeSuffix            = "";
        } else {
            unqualifiedTopLevelTypeName = className.substring(idx1, idx2);
            nestedTypeSuffix            = className.substring(idx2).replace('$', '.');
        }

        return new String[] { packagePrefix, unqualifiedTopLevelTypeName, nestedTypeSuffix };
    }

    /**
     * @see #generateLogging(Level, String, String, InsnList, InsnList, InsnList, InsnList, MethodVisitor)
     */
    private InsnList
    generateLogging(
        String        loggerName,
        Level         logLevel,
        String        sourceMethodName,
        int           line,
        String        message,
        InsnList      pushParameters,
        MethodVisitor methodVisitor
    ) {
        return this.generateLogging(
            loggerName,
            logLevel,
            null,             // optionalPrefix
            sourceMethodName,
            line,
            message,
            pushParameters,
            null,             // optionalSuffix
            methodVisitor
        );
    }

    /**
     * Generates code as follows:
     * <pre>
     *     [ <i>optional-prefix</i> ]
     *     Helper.logp(
     *         <i>this-class-name</i>.<i>logger-field-name</i>, // logger
     *         Level.<i>log-level-name</i>,                     // level
     *         <i>source-class-name</i>,                        // sourceClassName
     *         <i>source-method-name</i>,                       // sourceMethodName
     *         <i>source</i>,                                   // source
     *         <i>lineNumber</i>,                               // lineNumber
     *         <i>message</i>,                                  // message
     *         <i>push-parameters</i>                           // parameters
     *     );
     *     [ <i>optional-suffix</i> ]
     * </pre>
     */
    private InsnList
    generateLogging(
        String             loggerName,
        Level              logLevel,
        @Nullable InsnList prefix,
        @Nullable String   sourceMethodName,
        int                lineNumber,
        String             message,
        InsnList           pushParameters,
        @Nullable InsnList suffix,
        MethodVisitor      methodVisitor
    ) {
        assert message != null;

        if (logLevel == Level.OFF) return LoggifyingClassAdapter.NO_INSNS;
        InsnList il = new InsnList();

        this.loggingCount++;

        // [ optional-prefix ]
        if (prefix != null) il.add(prefix);

        // Helper.logp(
        //     this-class-name.logger-field-name, // logger
        //     Level.log-level-name,              // level
        //     source-class-name,                 // sourceClassName
        //     source-method-name,                // sourceMethodName
        //     source,                            // source
        //     lineNumber,                        // lineNumber
        //     message,                           // message
        //     push-parameters                    // parameters
        // );

        // Logger logger
        String loggerFieldName = LoggifyingClassAdapter.loggerNameToFieldName(loggerName);
        il.add(new FieldInsnNode(Opcodes.GETSTATIC, this.thisClassName, loggerFieldName, "Ljava/util/logging/Logger;"));
        // Level level
        il.add(LoggifyingClassAdapter.generatePushLogLevel(logLevel));
        // String sourceClassName
        il.add(
            this.sourceClassName == null
            ? new InsnNode(Opcodes.ACONST_NULL)
            : new LdcInsnNode(this.sourceClassName)
        );
        // String sourceMethodName
        il.add(sourceMethodName == null ? new InsnNode(Opcodes.ACONST_NULL) : new LdcInsnNode(sourceMethodName));
        // String source
        il.add(this.source == null ? new InsnNode(Opcodes.ACONST_NULL) : new LdcInsnNode(this.source));
        // int  lineNumber
        il.add(new LdcInsnNode(lineNumber));
        // String message
        il.add(new LdcInsnNode(message));
        // Object[] parameters
        il.add(pushParameters);

        il.add(new MethodInsnNode(Opcodes.INVOKESTATIC, LoggifyingClassAdapter.HELPER_CLASS_NAME, "logp", (
            "("
            + "Ljava/util/logging/Logger;"  // Logger   logger
            + "Ljava/util/logging/Level;"   // Level    level
            + "Ljava/lang/String;"          // String   sourceClassName
            + "Ljava/lang/String;"          // String   sourceMethodName
            + "Ljava/lang/String;"          // String   source
            + "I"                           // int      lineNumber
            + "Ljava/lang/String;"          // String   message
            + "[Ljava/lang/Object;"         // Object[] parameters
            + ")V"
        )));

        // [ optional-suffix ]
        if (suffix != null) il.add(suffix);

        return this.generateIsLoggable(loggerName, logLevel, il, null);
    }

    /**
     * Generates code as follows:
     * <pre>
     *     [ <i>optional-prefix</i> ]
     *     Helper.logp(
     *         <i>this-class-name</i>.<i>logger-field-name</i>,
     *         Level.<i>log-level-name</i>,
     *         <i>source-class-name</i>,
     *         <i>source-method-name</i>,
     *         <i>source</i>,
     *         <i>line-number</i>,
     *         <i>message</i>,
     *         <i>push-throwable</i>
     *     );
     *     [ <i>optional-suffix</i> ]
     * </pre>
     */
    private InsnList
    generateLoggingWithThrowable(
        String             loggerName,
        Level              logLevel,
        @Nullable InsnList prefix,
        @Nullable String   sourceMethodName,
        int                line,
        String             message,
        InsnList           pushThrowable,
        @Nullable InsnList suffix,
        MethodVisitor      methodVisitor
    ) {
        assert message != null;

        if (logLevel == Level.OFF) return LoggifyingClassAdapter.NO_INSNS;
        InsnList il = new InsnList();

        // [ optional-prefix ]
        if (prefix != null) il.add(prefix);

        // Helper.logp(
        //     this-class-name.logger-field-name, // logger
        //     Level.log-level-name,              // level
        //     source-class-name,                 // sourceClassName
        //     source-method-name,                // sourceMethodName
        //     source,                            // source
        //     line-number,                       // lineNumber
        //     message,                           // message
        //     push-throwable                     // throwable
        // );

        // Logger logger
        String loggerFieldName = LoggifyingClassAdapter.loggerNameToFieldName(loggerName);
        il.add(new FieldInsnNode(Opcodes.GETSTATIC, this.thisClassName, loggerFieldName, "Ljava/util/logging/Logger;"));
        // Level level
        il.add(LoggifyingClassAdapter.generatePushLogLevel(logLevel));
        // String sourceClassName
        il.add(
            this.sourceClassName == null
            ? new InsnNode(Opcodes.ACONST_NULL)
            : new LdcInsnNode(this.sourceClassName)
        );
        // String sourceMethodName
        il.add(sourceMethodName == null ? new InsnNode(Opcodes.ACONST_NULL) : new LdcInsnNode(sourceMethodName));
        // String source
        il.add(this.source == null ? new InsnNode(Opcodes.ACONST_NULL) : new LdcInsnNode(this.source));
        // int line
        il.add(new LdcInsnNode(line));
        // String message
        il.add(new LdcInsnNode(message));
        // Throwable thrown
        il.add(pushThrowable);

        il.add(new MethodInsnNode(Opcodes.INVOKESTATIC, LoggifyingClassAdapter.HELPER_CLASS_NAME, "logp", (
            "("
            + "Ljava/util/logging/Logger;" // logger
            + "Ljava/util/logging/Level;"  // level
            + "Ljava/lang/String;"         // sourceClassName
            + "Ljava/lang/String;"         // sourceMethodName
            + "Ljava/lang/String;"         // source
            + "I"                          // lineNumber
            + "Ljava/lang/String;"         // message
            + "Ljava/lang/Throwable;"      // thrown
            + ")V"
        )));

        // [ optional-suffix ]
        if (suffix != null) il.add(suffix);

        return this.generateIsLoggable(loggerName, logLevel, il, null);
    }

    /**
     * Generates code as follows:
     * <pre>
     * private static final ThreadLocal inLogging; // One JVM-global object.
     *
     * // ...
     *
     * if (<i>this-class-name</i>.<i>logger-field-name</i>.isLoggable(Level.<i>log-level-name</i>)) {
     *     <i>insns1</i>
     * } else {
     *     <i>optionalInsns2</i>
     * }
     * </pre>
     */
    private InsnList
    generateIsLoggable(String loggerName, Level logLevel, InsnList insns1, @Nullable InsnList insns2) {
        InsnList il              = new InsnList();
        String   loggerFieldName = LoggifyingClassAdapter.loggerNameToFieldName(loggerName);

        // if (this-class-name.logger-field-name.isLoggable(Level.log-level-name)) {
        LabelNode notLoggable = new LabelNode();
        il.add(new FieldInsnNode(Opcodes.GETSTATIC, this.thisClassName, loggerFieldName, "Ljava/util/logging/Logger;"));
        il.add(LoggifyingClassAdapter.generatePushLogLevel(logLevel));
        il.add(new MethodInsnNode(
            Opcodes.INVOKEVIRTUAL,
            "java/util/logging/Logger",
            "isLoggable",
            "(Ljava/util/logging/Level;)Z"
        ));
        il.add(new JumpInsnNode(Opcodes.IFEQ, notLoggable));

        //     <insns1>
        il.add(insns1);

        if (insns2 == null) {
            // }
            il.add(notLoggable);
        } else {
            // } else {
            LabelNode end = new LabelNode();
            il.add(new JumpInsnNode(Opcodes.GOTO, end));
            il.add(notLoggable);
            //     <insns2>
            il.add(insns2);
            // }
            il.add(end);
        }

        this.loggerNames.add(loggerName);
        return il;
    }

    private static AbstractInsnNode
    generatePushLogLevel(Level logLevel) {
        if (
            logLevel == Level.OFF
            || logLevel == Level.SEVERE
            || logLevel == Level.WARNING
            || logLevel == Level.INFO
            || logLevel == Level.CONFIG
            || logLevel == Level.FINE
            || logLevel == Level.FINER
            || logLevel == Level.FINEST
            || logLevel == Level.ALL
        ) {
            // Level.<log-level-name>
            return new FieldInsnNode(
                Opcodes.GETSTATIC,
                "java/util/logging/Level",
                logLevel.getName(),
                "Ljava/util/logging/Level;"
            );
        } else {
            // Level.parse("<log-level-name>")
            return new MethodInsnNode(
                Opcodes.INVOKESTATIC,
                "java/util/logging/Level",
                "parse",
                "(Ljava/lang/String;)Ljava/util/logging/Level;"
            );
        }
    }

    /**
     * Some logging systems don't like particular special characters in logger names, so replace them.
     */
    private static String
    cookLoggerName(String name) {
        return name.replace(' ', '_').replaceAll(",", "");
    }

    interface CurrentLineProvider {

        /** @return The related line number, which is 1 or greater */
        int getCurrentLine();
    }

    interface LocalVariableFinder {

        /** @return The {@link LocalVariableNode} related to the given local variable {@code index} */
        @Nullable LocalVariableNode getLocalVariable(int index);
    }

    /**
     * @return A human-readable string representing the given {@code access}, for classes (i.e. 0x20 is SUPER, not
     *         SYNCHRONIZED), followed by a SPACE
     */
    public static String
    classAccessToString(int access) {
        return LoggifyingClassAdapter.accessToString(access, LoggifyingClassAdapter.CLASS_ACCESSES);
    }

    /**
     * @return A human-readable string representing the given {@code access}, for fields (i.e. 0x20 is SYNCHRONIZED, not
     *         SUPER, and 0x40 is VOLATILE, not BRDIGE, and 0x80 is TRANSIENT, not VARARGS), followed by a SPACE
     */
    public static String
    fieldAccessToString(int access) {
        return LoggifyingClassAdapter.accessToString(access, LoggifyingClassAdapter.FIELD_ACCESSES);
    }

    /**
     * @return A human-readable string representing the given {@code access}, for methods (i.e. 0x20 is SYNCHRONIZED,
     *         not SUPER, and 0x40 is BRDIGE, not VOLATILE, and 0x80 is VARARGS, not TRANSIENT), followed by a SPACE
     */
    public static String
    methodAccessToString(int access) {
        return LoggifyingClassAdapter.accessToString(access, LoggifyingClassAdapter.METHOD_ACCESSES);
    }

    private static String
    accessToString(int access, Object[] table) {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < table.length;) {
            int    a = (Integer) table[i++];
            String s = (String) table[i++];

            if ((access & a) != 0) sb.append(s);
        }
        return sb.length() > 0 ? sb.substring(0, sb.length() - 1) : "";
    }
    private static final Object[] CLASS_ACCESSES = {
        Opcodes.ACC_PUBLIC,     "public ",
        Opcodes.ACC_PRIVATE,    "private ",
        Opcodes.ACC_PROTECTED,  "protected ",
        Opcodes.ACC_FINAL,      "final ",
        Opcodes.ACC_SUPER,      "super ",
        Opcodes.ACC_INTERFACE,  "interface ",
        Opcodes.ACC_ABSTRACT,   "abstract ",
        Opcodes.ACC_SYNTHETIC,  "synthetic ",
        Opcodes.ACC_ANNOTATION, "annotation ",
        Opcodes.ACC_ENUM,       "enum ",
        Opcodes.ACC_DEPRECATED, "deprecated ",
    };
    private static final Object[] FIELD_ACCESSES = {
        Opcodes.ACC_PUBLIC,     "public ",
        Opcodes.ACC_PRIVATE,    "private ",
        Opcodes.ACC_PROTECTED,  "protected ",
        Opcodes.ACC_FINAL,      "final ",
        Opcodes.ACC_STATIC,     "static ",
        Opcodes.ACC_SUPER,      "synchronized ", // sonst SUPER
        Opcodes.ACC_VOLATILE,   "volatile ",
        Opcodes.ACC_TRANSIENT,  "transient ",
        Opcodes.ACC_SYNTHETIC,  "synthetic ",
        Opcodes.ACC_ENUM,       "enum ",
        Opcodes.ACC_DEPRECATED, "deprecated ",
    };
    private static final Object[] METHOD_ACCESSES = {
        Opcodes.ACC_PUBLIC,       "public ",
        Opcodes.ACC_PRIVATE,      "private ",
        Opcodes.ACC_PROTECTED,    "protected ",
        Opcodes.ACC_FINAL,        "final ",
        Opcodes.ACC_STATIC,       "static ",
        Opcodes.ACC_SYNCHRONIZED, "synchronized ",
        Opcodes.ACC_BRIDGE,       "bridge ",
        Opcodes.ACC_VARARGS,      "varargs ",
        Opcodes.ACC_NATIVE,       "native ",
        Opcodes.ACC_STRICT,       "strict ",
        Opcodes.ACC_ABSTRACT,     "abstract ",
        Opcodes.ACC_SYNTHETIC,    "synthetic ",
        Opcodes.ACC_DEPRECATED,   "deprecated ",
    };

    /** Handles various error conditions which the LOGGIFIER runs. */
    public
    interface ErrorHandler {

        /** Handles an error with the given {@code message}. */
        void
        handle(String message);

        /** Handles an error with the given {@link Throwable}. */
        void
        handle(Throwable t);

        /** An {@link ErrorHandler} which ignores any errors. */
        ErrorHandler IGNORE = new ErrorHandler() {

            @Override public void
            handle(String message) {}

            @Override public void
            handle(Throwable t) {}
        };

        /** An {@link ErrorHandler} which prints errors to STDERR. */
        ErrorHandler STDERR = new ErrorHandler() {

            @Override public void
            handle(String message) { System.err.println(message); }

            @Override public void
            handle(Throwable t) { t.printStackTrace(); }
        };
    }
}
