
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.logging.Level;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Opcodes;

import de.unkrig.commons.file.contentstransformation.ContentsTransformer;
import de.unkrig.commons.file.contentstransformation.ContentsTransformerUtil;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.loggifier.LoggifyingClassAdapter.ErrorHandler;
import de.unkrig.loggifier.loglevel.LevelCalculator;

/**
 * A {@link ContentsTransformer} that loggifies a class file, and copies the contents of any other file unchanged.
 */
public
class LoggifyingContentsTransformer implements ContentsTransformer {

    /**
     * The 'actions' that the byte code of a method might execute. Appear in alphabetical order for convenience.
     */
    public
    enum Action {

        /**
         * Logs array type, array value, index and element value read from an array"; }
         * @see #ASTORE
         */
        ALOAD {
            @Override public String
            description() {
                return "Log array type, array value, index and element value read from an array";
            }
        },

        /**
         * Logs operands, operator (+, -, *, /, %, &amp;, |, ^, &lt;=> (comparison), ==, !=, &lt;, >=, >, &lt;=,
         * unary -, == null, != null, &lt;&lt;, >>, >>>) and result of an arithmetic operation.
         */
        ARITH {
            @Override public String
            description() {
                return "Log operands, operator (+, -, ...) and result of an arithmetic operation";
            }
        },

        /**
         * Logs the array, the index and the value being stored in an array.
         * @see #ALOAD
         */
        ASTORE {
            @Override public String
            description() {
                return "Log the array, the index and the value being stored in an array";
            }
        },

        /**
         * Logs the type to which the operand is casted.
         */
        CAST {
            @Override public String
            description() {
                return "Log the type to which the operand is casted";
            }
        },

        /**
         * Logs the declared catch type and the actually caught exception.
         * @see #THROW
         */
        CATCH {
            @Override public String
            description() {
                return "Log the declared catch type and the actually caught exception";
            }
        },

        /**
         * Logs the name of a class that is initialized.
         */
        CLINIT {
            @Override public String
            description() {
                return "Log the name of a class that is initialized";
            }
        },

        /**
         * Logs the value of a constant being pushed on the operand stack.
         */
        CONST {
            @Override public String
            description() {
                return "Log the value of a constant being pushed on the operand stack";
            }
        },

        /**
         * Logs a value that is arithmetically converted and the type of the conversion result.
         */
        CONVERT {
            @Override public String
            description() {
                return "Log a value that is arithmetically converted and the conversion result";
            }
        },

        /**
         * Logs the target object (if non-static) and the arguments as a method begins execution.
         * @see #RETURN
         */
        ENTRY {
            @Override public String
            description() {
                return "Log the target object (if non-static) and the arguments as a method begins execution";
            }
        },

        /**
         * Logs declaring class of, name of, source instance of (if non-static) and value read from a field.
         * @see #PUT
         */
        GET {
            @Override public String
            description() {
                return "Log declaring class of, name of, source instance of (if non-static) and value read from a field"; // SUPPRESS CHECKSTYLE LineLength
            }
        },

        /**
         * Logs signature, target object (if non-static) and arguments right before a method is invoked.
         * @see Loggifier#RESULT
         */
        INVOKE {
            @Override public String
            description() {
                return "Log signature, target object (if non-static) and arguments right before a method is invoked";
            }
        },

        /**
         * Logs the type to which the operand type is compared.
         */
        INSTANCEOF {
            @Override public String
            description() {
                return "Log the type to which the operand type is compared";
            }
        },

        /**
         * Logs the array who's length is determined.
         */
        LENGTH {
            @Override public String
            description() {
                return "Log the array who's length is determined";
            }
        },

        /**
         * Logs type of, name of and value read from a local variable.
         * @see #STORE
         */
        LOAD {
            @Override public String
            description() {
                return "Log type of, name of and value read from a local variable";
            }
        },

        /**
         * Logs the type of a new object being instantiated.
         */
        NEW {
            @Override public String
            description() {
                return "Log the type of a new object being instantiated";
            }
        },

        /**
         * Logs declaring class of, name of, target instance of (if non-static) and value assigned to field.
         * @see #GET
         */
        PUT {
            @Override public String
            description() {
                return "Log declaring class of, name of, target instance of (if non-static) and value assigned to field"; // SUPPRESS CHECKSTYLE LineLength
            }
        },

        /**
         * Logs method signature and invocation result (if not 'void').
         * @see Loggifier#INVOKE
         */
        RESULT {
            @Override public String
            description() {
                return "Log method signature and invocation result (if not 'void')";
            }
        },

        /**
         * Logs the value that is being returned by a method (if not 'void').
         * @see #ENTRY
         */
        RETURN {
            @Override public String
            description() {
                return "Log the value that is being returned by a method (if not 'void')";
            }
        },

        /**
         * Logs value assigned to, type of and name of a local variable.
         * @see #LOAD
         */
        STORE {
            @Override public String
            description() {
                return "Log value assigned to, type of and name of a local variable";
            }
        },

        /**
         * Logs the integer value that a switch is being executed on.
         */
        SWITCH {
            @Override public String
            description() {
                return "Log the integer value that a switch is being executed on";
            }
        },

        /**
         * Logs an exception right before it is thrown.
         * @see #CATCH
         */
        THROW {
            @Override public String
            description() {
                return "Log an exception right before it is thrown";
            }
        };

        /** @return A human-readable, english description of this action */
        public abstract String description();
    }

    /**
     * Calculates the default loggification {@link Level} for the given {@link Action}.
     */
    public static final LevelCalculator DEFAULT_LOG_LEVEL_CALCULATOR = new LevelCalculator() {

        @Override public Level
        calculate(
            Action           action,
            @Nullable String source,
            int              classAccess,
            String           className,
            int              methodAccess,
            String           methodName,
            int              lineNumber,
            @Nullable String custom
        ) {

            // Accesses to local variables and fields who's names end with '__', and instantiations of classes who's
            // names end with '__' are loggified!
            if (custom != null && custom.endsWith("__")) return Level.FINER;

            if (action == Action.ENTRY)  return Level.FINER;
            if (action == Action.RETURN) return Level.FINER;
            if (action == Action.THROW)  return Level.FINER;

            return Level.OFF;
        }
    };

    /***/
    public static final Level
    getDefaultLevel(Action action) {
        return LoggifyingContentsTransformer.DEFAULT_LOG_LEVEL_CALCULATOR.calculate(
            action,
            "",                 // source
            Opcodes.ACC_PUBLIC, // classAccess
            "",                 // className
            Opcodes.ACC_PUBLIC, // methodAccess
            "",                 // methodName
            1,                  // lineNumber
            null                // custom
        );
    }

    private final LevelCalculator logLevelCalculator;
    private final ErrorHandler    errorHandler;

    public
    LoggifyingContentsTransformer(LevelCalculator logLevelCalculator, ErrorHandler errorHandler) {
        this.logLevelCalculator = logLevelCalculator;
        this.errorHandler       = errorHandler;
    }

    /**
     * @throws ConstantPoolTooLargeException
     * @throws ExceptionTableEntryOffsetTooLargeException
     */
    @Override public void
    transform(String name, InputStream is, OutputStream os) throws IOException {

        byte[] bytecode;
        try {
            bytecode = this.transform(name, is);
        } catch (DoubleLoggificationException dle) {
            is.reset();
            ContentsTransformerUtil.copy().transform(name, is, os);
            return;
        }
        if (bytecode == null) {

            // Not a class file, copy contents unchanged.
            ContentsTransformerUtil.copy().transform(name, is, os);
            return;
        }

        new DataOutputStream(os).write(bytecode);
    }

    /**
     * @return                                            {@code null} iff the {@link InputStream} is not in class file
     *                                                    format
     * @throws MethodCodeTooLargeException
     * @throws ConstantPoolTooLargeException
     * @throws ExceptionTableEntryOffsetTooLargeException
     */
    @Nullable public byte[]
    transform(final String name, InputStream is) throws IOException {

        // Make <is> resettable, because we need to read the magic number at the beginning of a class file.
        if (!is.markSupported()) {
            is = new BufferedInputStream(is, 4);
        }

        // Attempt to read the magic number.
        {
            is.mark(4);
            int magic;
            try {
                magic = new DataInputStream(is).readInt();
            } catch (EOFException eofe) {
                magic = 0;
            }
            is.reset();
            if (magic != 0xcafebabe) {

                // Not a class file.
                return null;
            }
        }

        // Read, loggify and write the CLASS file.
        ClassReader            cr  = new ClassReader(is);
        ClassWriter            cw  = new ClassWriter(cr, ClassWriter.COMPUTE_FRAMES);
        LoggifyingClassAdapter lca = new LoggifyingClassAdapter(cw, this.logLevelCalculator, this.errorHandler);
        cr.accept(lca, 0);
        byte[] bytecode;
        try {
            bytecode = cw.toByteArray();
        } catch (RuntimeException re) {

            // ASM 4.0, MethodWriter.java, line 1856.
            if ("Method code too large!".equals(re.getMessage())) {
                // SUPPRESS CHECKSTYLE AvoidHidingCause
                throw new MethodCodeTooLargeException("Please tighten the loggification rules for this class");
            }
            throw re;
        }

        // "ClassWriter.toByteArray()" does not check for constant pool overflow - need to check ourselves...
        if (cw.newUTF8("sldkfhslkdfhl") >= 65536) {
            throw new ConstantPoolTooLargeException("Please tighten the loggification rules for this class");
        }

        return bytecode;
    }
}
