
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import de.unkrig.commons.file.filetransformation.FileTransformationUtil;
import de.unkrig.commons.lang.protocol.Predicate;
import de.unkrig.commons.lang.protocol.PredicateUtil;
import de.unkrig.commons.text.pattern.Glob;
import de.unkrig.commons.text.pattern.Pattern2;
import de.unkrig.commons.util.logging.SimpleLogging;
import de.unkrig.loggifier.LoggifyingClassAdapter.ErrorHandler;
import de.unkrig.loggifier.LoggifyingContentsTransformer.Action;
import de.unkrig.loggifier.RuleSet.RuleException;

/** The LOGGIFIER command line tool. */
public final
class Main {

    private Main() {}

    static { SimpleLogging.init(); }

    /**/
    public static void
    main(String[] args) throws IOException {
        boolean           keepOriginals = false;
        Predicate<String> selector      = PredicateUtil.always();
        RuleSet           ruleSet       = new RuleSet(LoggifyingContentsTransformer.DEFAULT_LOG_LEVEL_CALCULATOR);

        int i = 0;
        while (i < args.length) {
            String arg = args[i];
            if (!arg.startsWith("-")) break;
            i++;
            if ("-".equals(arg)) break;

            if ("-help".equals(arg)) {

                String actions1, actions2, actions3;
                {
                    actions1 = Arrays.toString(Action.values());
                    int idx = actions1.lastIndexOf(", ", 57);
                    actions2 = actions1.substring(idx + 2);
                    actions1 = actions1.substring(0, idx + 1);

                    idx      = actions2.lastIndexOf(", ", 65);
                    actions3 = actions2.substring(idx + 2);
                    actions2 = actions2.substring(0, idx + 1);
                }

                System.out.printf(
                    ""
                    + "Implants logging code into Java class files 'in-place', also in ZIP format%n"
                    + "files (zip, jar, war, ear, ...) and nested ZIP format files.%n"
                    + "%n"
                    + "Usage:%n"
                    + "  java " + Main.class.getName() + " [ <option> ] ... <dir-or-file>%n"
                    + "  java " + Main.class.getName() + " [ <option> ] ... <dir-or-file> <dir-or-file>%n"
                    + "  java " + Main.class.getName() + " [ <option> ] ... <dir-or-file> ... <dir>%n"
                    + "%n"
                    + "Valid options:%n"
                    + "  -help           Shows this text%n"
                    + "  -keep           Keep copies of original files%n"
                    + "  -selector <glob>%n"
                    + "                  Loggify only classes matching the <glob>%n"
                    + "  -rule <rule>    Add the given rule%n"
                    + "  -nowarn         Suppress all messages except errors%n"
                    + "  -quiet          Suppress normal output%n"
                    + "  -verbose        Log verbose messages%n"
                    + "  -debug          Log verbose and debug messages%n"
                    + "  -log <level>:<logger>:<handler>:<formatter>:<format>%n"
                    + "                  Add logging at level FINE on logger 'de.unkrig' to STDERR%n"
                    + "                  using the FormatFormatter and SIMPLE format, or the given%n"
                    + "                  arguments which are all optional.%n"
                    + "The rule format is as follows%n"
                    + "  <action>[,<action>]...=<level>[:<expression>]%n"
                    + "%n"
                    + "<action>:     One of " + actions1 + "%n"
                    + "              " + actions2 + "%n"
                    + "              " + actions3 + "%n"
                    + "<level>:      One of [OFF SEVERE WARNING INFO CONFIG FINE FINER FINEST]%n"
                    + "<expression>: E.g. '(<eq> || <eq>) && !<eq>', where%n"
                    + "<eq>:         '<var> == <glob>', where%n"
                    + "<var>:        is one of [action source classAccess class methodAccess method%n"
                    + "              line type classAndField variable].%n"
                    + "              For actions [CAST INSTANCEOF NEW] also 'type'.%n"
                    + "              For actions [GET PUT] also 'classAndField'.%n"
                    + "              For actions [LOAD STORE] also 'name'.%n"
                    + "<glob>:       is a 'glob pattern' in single quotes like '*main*'.%n"
                    + "%n"
                    + "The exit status is 1 iff errors occurred during loggification.%n"
                );
                System.exit(0);
            } else
            if ("-keep".equals(arg)) {
                keepOriginals = true;
            } else
            if ("-selector".equals(arg)) {
                selector = Glob.compile(args[i++], Pattern2.WILDCARD | Glob.INCLUDES_EXCLUDES);
            } else
            if ("-rule".equals(arg)) {
                try {
                    ruleSet.add(args[i++]);
                } catch (RuleException re) {
                    System.err.println(re.getMessage());
                    System.exit(1);
                }
            } else
            if ("-nowarn".equals(arg)) {
                SimpleLogging.setNoWarn();
            } else
            if ("-quiet".equals(arg)) {
                SimpleLogging.setQuiet();
            } else
            if ("-verbose".equals(arg)) {
                SimpleLogging.setVerbose();
            } else
            if ("-debug".equals(arg)) {
                SimpleLogging.setDebug();
            } else
            if ("-log".equals(arg)) {
                SimpleLogging.configureLoggers(args[i++]);
            } else
            {
                System.err.println("Unknown command line option '" + arg + "', try '-help'.");
                System.exit(1);
            }
        }

        List<File> inputFiles;
        File       outputFile;

        if (i == args.length) {
            System.err.println("Input file name missing, try '-help'.");
            System.exit(1);
            return;
        } else
        if (i == args.length - 1) {
            inputFiles = Collections.singletonList(new File(args[i]));
            outputFile = null;
        } else
        {
            inputFiles = new ArrayList<File>();
            for (; i < args.length - 1; i++) inputFiles.add(new File(args[i]));
            outputFile = new File(args[i]);
        }

        // Apply the file transformer to all input files.
        final int[] errorCount = new int[1];
        Loggifier   loggifier  = new Loggifier(selector, ruleSet, keepOriginals, new ErrorHandler() {

            @Override public void
            handle(String message) {
                System.err.println(message);
                errorCount[0]++;
            }

            @Override public void
            handle(Throwable t) {
                t.printStackTrace();
                errorCount[0]++;
            }
        });
        FileTransformationUtil.transform(inputFiles, outputFile, loggifier);

        if (errorCount[0] > 0) {
            System.err.println(errorCount[0] + " error(s) occurred during loggification.");
            System.exit(1);
        }
    }
}
