
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import java.util.EnumSet;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.unkrig.commons.lang.ExceptionUtil;
import de.unkrig.commons.nullanalysis.NotNull;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.commons.text.parser.ParseException;
import de.unkrig.loggifier.LoggifyingContentsTransformer.Action;
import de.unkrig.loggifier.loglevel.LevelCalculator;
import de.unkrig.loggifier.loglevel.LevelCalculatorUtil;

/**
 * @see #calculate(Action, String, int, String, int, String, int, String)
 */
public
class RuleSet implements LevelCalculator {
    private static final Pattern RULE_PATTERN = java.util.regex.Pattern.compile("([^=]+)=([^:]+)(?::(.*))?");
    private LevelCalculator      delegate;

    /** Representation of a problem that occurred when a rule was parsed. */
    public static
    class RuleException extends Exception {

        private static final long serialVersionUID = 1L;

        public
        RuleException(String message) {
            super(message);
        }

        public
        RuleException(String message, Throwable t) {
            super(message, t);
        }

        public
        RuleException(Throwable t) {
            super(t);
        }
    }

    /**
     * @see #calculate(Action, String, int, String, int, String, int, String)
     */
    public
    RuleSet(LevelCalculator parent) {
        this.delegate = parent;
    }

    /**
     * Applies the given values to one rule after another, and returns the log level of the first matching rule. Iff
     * none of the rules apply, the {@link #RuleSet(LevelCalculator) parent} calculator is invoked and its result
     * returned.
     */
    @Override public Level
    calculate(
        Action           action,
        @Nullable String source,
        int              classAccess,
        String           className,
        int              methodAccess,
        String           methodName,
        int              lineNumber,
        @Nullable String custom
    ) {
        return this.delegate.calculate(
            action,
            source,
            classAccess,
            className,
            methodAccess,
            methodName,
            lineNumber,
            custom
        );
    }

    /** Parses the given {@code rule} and adds it to this {@link RuleSet}. */
    public RuleSet
    add(String rule) throws RuleException {

        // Break up the rule.
        @NotNull String      actionName;
        @NotNull final Level level;
        @Nullable String     discriminatorExpression;
        {
            Matcher m = RuleSet.RULE_PATTERN.matcher(rule);
            if (!m.matches()) {
                throw new RuleException(
                    "Loggification rule '"
                    + rule
                    + "': Invalid format - must be like '<action>[,<action]...=<level>[:<expression>"
                );
            }
            actionName = m.group(1).toUpperCase();
            try {
                level = Level.parse(m.group(2));
            } catch (IllegalArgumentException iae) {
                throw new RuleException("Loggification rule '" + rule + "': " + iae.getMessage(), iae);
            }
            discriminatorExpression = m.group(3);
        }

        final EnumSet<Action> actions;
        if ("ALL".equals(actionName)) {
            actions = EnumSet.allOf(Action.class);
        } else {
            actions = EnumSet.noneOf(Action.class);
            for (StringTokenizer st = new StringTokenizer(actionName, ", "); st.hasMoreTokens();) {
                actions.add(Action.valueOf(st.nextToken()));
            }
        }

        try {
            this.delegate = LevelCalculatorUtil.discriminated(actions, level, discriminatorExpression, this.delegate);
        } catch (ParseException pe) {
            throw ExceptionUtil.wrap(
                "Parsing discriminator expression '" + discriminatorExpression + "'",
                pe,
                RuleException.class
            );
        }

        return this;
    }
}
