
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier.loglevel;

import java.util.logging.Level;

import org.objectweb.asm.Opcodes;

import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.loggifier.LoggifyingContentsTransformer.Action;

/**
 * @see #calculate(Action, String, int, String, int, String, int, String)
 */
public
interface LevelCalculator {

    /** A {@link LevelCalculator} that always returns {@link Level#OFF} */
    LevelCalculator ALWAYS = LevelCalculatorUtil.ALWAYS_OFF;

    /** A {@link LevelCalculator} that always returns {@link Level#ALL} */
    LevelCalculator NEVER = LevelCalculatorUtil.ALWAYS_ALL;

    /**
     * Calculates a logging level depending on the parameter values.
     * <p>
     * The meaning of the {@code custom} parameter depends on the value of the {@code action} argument:
     * <dl>
     *   <dt>CAST, INSTANCEOF:</dt>
     *   <dd>{@code type=pkg.Class, pkg.Class[][], int[][]}</dd>
     *   <dt>CAST, INSTANCEOF, NEW:</dt>
     *   <dd>{@code type=pkg.Class, pkg.Class[x], pkg.Class[x][]}</dd>
     *   <dt>GET, PUT:</dt>
     *   <dd>{@code classAndField=pkg.Class.field}</dd>
     *   <dt>LOAD, STORE:</dt>
     *   <dd>{@code variable=locvar}</dd>
     *   <dt>All other:</dt>
     *   <dd>null</dd>
     * </dl>
     *
     * @param action
     * @param source       The simple source file name (without directories)
     * @param classAccess  Any combination of the {@code ACC_*} constants in {@link Opcodes}
     * @param className    The fully qualified name of the class being instrumented, e.g. '{@code pkg.MyClass}'
     * @param methodAccess Any combination of the {@code ACC_*} constants in {@link Opcodes}
     * @param methodName   The enhanced, dequalified name of the method being instrumented, e.g. '{@code meth(String,
     *                     MyClass)}' or '{@code <T extends IOException>meth(T, MyClass, List<T>)}'
     * @param lineNumber
     * @param custom       See above
     */
    Level
    calculate(
        Action           action,
        @Nullable String source,
        int              classAccess,
        String           className,
        int              methodAccess,
        String           methodName,
        int              lineNumber,
        @Nullable String custom
    );
}
