
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2013, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier.loglevel;

import java.util.EnumSet;
import java.util.logging.Level;

import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.commons.text.expression.EvaluationException;
import de.unkrig.commons.text.expression.Expression;
import de.unkrig.commons.text.expression.ExpressionEvaluator;
import de.unkrig.commons.text.expression.ExpressionUtil;
import de.unkrig.commons.text.parser.ParseException;
import de.unkrig.loggifier.LoggifyingClassAdapter;
import de.unkrig.loggifier.LoggifyingContentsTransformer.Action;

/**
 * Various {@link LevelCalculator}-related utility methods.
 */
public final
class LevelCalculatorUtil {

    private
    LevelCalculatorUtil() {}

    /**
     * @return A {@link LevelCalculator} which always calculates the constant value {@code level}
     */
    public static LevelCalculator
    constant(final Level level) {
        return new LevelCalculator() {

            @Override public Level
            calculate(
                Action           action,
                @Nullable String source,
                int              classAccess,
                String           className,
                int              methodAccess,
                String           methodName,
                int              lineNumber,
                @Nullable String custom
            ) {
                return level;
            }

            @Override public String
            toString() {
                return level.toString();
            }
        };
    }

    /** A {@link LevelCalculator} which always calculates the constant value {@link Level#OFF}. */
    public static final LevelCalculator ALWAYS_OFF = LevelCalculatorUtil.constant(Level.OFF);

    /** A {@link LevelCalculator} which always calculates the constant value {@link Level#FINE}. */
    public static final LevelCalculator ALWAYS_FINE = LevelCalculatorUtil.constant(Level.FINE);

    /** A {@link LevelCalculator} which always calculates the constant value {@link Level#FINER}. */
    public static final LevelCalculator ALWAYS_FINER = LevelCalculatorUtil.constant(Level.FINER);

    /** A {@link LevelCalculator} which always calculates the constant value {@link Level#FINEST}. */
    public static final LevelCalculator ALWAYS_FINEST = LevelCalculatorUtil.constant(Level.FINEST);

    /** A {@link LevelCalculator} which always calculates the constant value {@link Level#ALL}. */
    public static final LevelCalculator ALWAYS_ALL = LevelCalculatorUtil.constant(Level.ALL);

    /**
     * Returns a {@link LevelCalculator} which returns the given {@code level} iff the action is contained in the
     * given {@code actions} and the {@code discriminatorExpression} (if any) evaluates to {@code true}.
     * Otherwise, or if the {@code discriminatorExpression} throws an {@link EvaluationException}, the {@code parent}
     * level calculator is invoked and the result returned.
     */
    public static LevelCalculator
    discriminated(
        final EnumSet<Action> actions,
        final Level           level,
        @Nullable String      discriminatorExpression,
        final LevelCalculator parent
    ) throws ParseException {

        return LevelCalculatorUtil.discriminated(
            actions,
            level,
            (
                discriminatorExpression == null
                ? ExpressionUtil.CONSTANTLY_TRUE
                : new ExpressionEvaluator(LevelCalculatorUtil.VARIABLE_NAMES).parse(discriminatorExpression)
            ),
            parent
        );
    }

    /**
     * Returns a {@link LevelCalculator} which returns the given {@code level} iff the action is contained in the
     * given {@code actions} and the {@code discriminatorExpression} evaluates to {@code true}.
     * Otherwise, or if the {@code discriminatorExpression} throws an {@link EvaluationException}, the {@code parent}
     * level calculator is invoked and the result returned.
     */
    public static LevelCalculator
    discriminated(
        final EnumSet<Action> actions,
        final Level           level,
        final Expression      discriminatorExpression,
        final LevelCalculator parent
    ) {

        return new LevelCalculator() {

            @Override public Level
            calculate(
                final Action           action,
                @Nullable final String source,
                final int              classAccess,
                final String           className,
                final int              methodAccess,
                final String           methodName,
                final int              lineNumber,
                @Nullable final String custom
            ) {
                try {

                    boolean type          = LevelCalculatorUtil.ACTIONS_WITH_TYPE.contains(action);
                    boolean classAndField = LevelCalculatorUtil.ACTIONS_WITH_CLASS_AND_FIELD.contains(action);
                    boolean variable      = LevelCalculatorUtil.ACTIONS_WITH_VARIABLE.contains(action);

                    if (actions.contains(action) && ExpressionUtil.evaluateTo(
                        discriminatorExpression,
                        new Object[] {
                            action,                                                    // action
                            source,                                                    // source
                            LoggifyingClassAdapter.classAccessToString(classAccess),   // classAccess
                            className,                                                 // class
                            LoggifyingClassAdapter.methodAccessToString(methodAccess), // methodAccess
                            methodName,                                                // method
                            lineNumber,                                                // line
                            type            ? custom : null,                           // type
                            classAndField ? custom : null,                             // classAndField
                            variable        ? custom : null,                           // variable
                        },
                        Boolean.class
                    )) return level;
                } catch (EvaluationException e) {
                    ;
                }
                return parent.calculate(
                    action,
                    source,
                    classAccess,
                    className,
                    methodAccess,
                    methodName,
                    lineNumber,
                    custom
                );
            }
        };
    }

    private static final String[] VARIABLE_NAMES = {
        "action",
        "source",
        "classAccess",
        "class",
        "methodAccess",
        "method",
        "line",
        "type",
        "classAndField",
        "variable",
    };
    private static final EnumSet<Action>
    ACTIONS_WITH_TYPE            = EnumSet.<Action>of(Action.CAST, Action.INSTANCEOF, Action.NEW);
    private static final EnumSet<Action>
    ACTIONS_WITH_CLASS_AND_FIELD = EnumSet.<Action>of(Action.GET, Action.PUT);
    private static final EnumSet<Action>
    ACTIONS_WITH_VARIABLE        = EnumSet.<Action>of(Action.LOAD, Action.STORE);
}
