
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier.runtime;

import java.util.logging.Level;
import java.util.logging.Logger;

/** Utility methods which are used by loggified code. */
public final
class Helper {

    private
    Helper() {}

    /**
     * Non-null while we are logging, in order to prevent endless recursion through objects' loggified 'toString()'
     * methods.
     */
    private static final ThreadLocal<Object> IN_LOGGING = new ThreadLocal<Object>();

    /**
     * A wrapper around {@link Logger#logp(Level, String, String, String, Object[])} which does some formatting and
     * pretty-printing.
     */
    public static void
    logp(
        Logger   logger,
        Level    level,
        String   sourceClassName,
        String   sourceMethodName,
        String   source,
        int      lineNumber,
        String   message,
        Object[] parameters
    ) {
        if (Helper.IN_LOGGING.get() == null) {
            Helper.IN_LOGGING.set(""); // To prevent endless recursion.
            try {

                // Wrap the parameters in a "pretty printer" object, so that CHARs, CharSequences and arrays are nicely
                // formatted.
                if (parameters != null) {
                    for (int i = 0; i < parameters.length; i++) {
                        final Object argument = parameters[i];
                        if (argument != null) {
                            Class<? extends Object> argumentClass = argument.getClass();
                            if (
                                argumentClass.isArray()
                                || argument instanceof CharSequence
                                || argumentClass == Character.class
                                || argumentClass == Byte.class
                                || argumentClass == Short.class
                                || argumentClass == Long.class
                                || argumentClass == Float.class
                                || argumentClass == Double.class
                            ) {

                                // Wrap the argument in an Object to delay the string conversion until
                                // 'MessageFormat.format()' is invoked, which sometimes never happens, e.g. if
                                // the log record is filtered out.
                                parameters[i] = new Object() {

                                    @Override public String
                                    toString() {
                                        return PrettyPrinter.toString(argument);
                                    }
                                };
                            }
                        }
                    }
                }

                logger.logp(
                    level,
                    sourceClassName,
                    sourceMethodName,
                    (                          // message
                        new StringBuilder("(")
                        .append(source)
                        .append(':')
                        .append(lineNumber)
                        .append(") ")
                        .append(message)
                        .toString()
                    ),
                    parameters
                );
            } finally {
                Helper.IN_LOGGING.set(null); // Much cheaper than 'ThreadLocal.remove()'.
            }
        }
    }

    /**
     * A wrapper around {@link Logger#logp(Level, String, String, String, Throwable)} which does some formatting and
     * pretty-printing.
     */
    public static void
    logp(
        Logger    logger,
        Level     level,
        String    sourceClassName,
        String    sourceMethodName,
        String    source,
        int       lineNumber,
        String    message,
        Throwable thrown
    ) {
        if (Helper.IN_LOGGING.get() == null) {
            Helper.IN_LOGGING.set(""); // To prevent endless recursion.
            try {
                logger.logp(
                    level,
                    sourceClassName,
                    sourceMethodName,
                    (                          // message
                        new StringBuilder("(")
                        .append(source)
                        .append(':')
                        .append(lineNumber)
                        .append(") ")
                        .append(message)
                        .toString()
                    ),
                    thrown
                );
            } finally {
                Helper.IN_LOGGING.set(null); // Much cheaper than 'ThreadLocal.remove()'.
            }
        }
    }
}
